#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core import debug
from fenrirscreenreader.core.screenDriver import ScreenDriver as screenDriver


class driver(screenDriver):
    def __init__(self):
        screenDriver.__init__(self)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core import debug


class speech_driver:
    def __init__(self):
        pass

    def initialize(self, environment):
        self._is_initialized = False
        self.language = None
        self.voice = None
        self.module = None
        self.pitch = None
        self.rate = None
        self.volume = None
        self.env = environment
        self._is_initialized = True

    def shutdown(self):
        if self._is_initialized:
            self.cancel()
        self._is_initialized = False

    def speak(self, text, queueable=True, ignore_punctuation=False):
        if not self._is_initialized:
            return
        if not queueable:
            self.cancel()

    def cancel(self):
        if not self._is_initialized:
            return

    def set_callback(self, callback):
        if not self._is_initialized:
            return
        if not callback:
            return

    def clear_buffer(self):
        if not self._is_initialized:
            return

    def set_voice(self, voice):
        if not self._is_initialized:
            return
        if voice == "":
            return
        self.voice = voice

    def set_pitch(self, pitch):
        if not self._is_initialized:
            return
        if not isinstance(pitch, float):
            return
        if pitch < 0.0:
            return
        if pitch > 1.0:
            return
        self.pitch = pitch

    def set_rate(self, rate):
        if not self._is_initialized:
            return
        if not isinstance(rate, float):
            return
        if rate < 0.0:
            return
        if rate > 1.0:
            return
        self.rate = rate

    def set_module(self, module):
        if not self._is_initialized:
            return
        if not isinstance(module, str):
            return
        if module == "":
            return
        self.module = module

    def reset(self):
        self.shutdown()
        self.initialize(self.env)

    def set_language(self, language):
        if not self._is_initialized:
            return
        if not isinstance(language, str):
            return
        if language == "":
            return
        self.language = language

    def set_volume(self, volume):
        if not self._is_initialized:
            return
        if not isinstance(volume, float):
            return
        if volume < 0.0:
            return
        if volume > 1.0:
            return
        self.volume = volume



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from collections import Counter

from fenrirscreenreader.core import debug
from fenrirscreenreader.core.i18n import _


class AttributeManager:
    def __init__(self):
        self.currAttributes = None
        self.prevAttributes = None
        self.currAttributeDelta = ""
        self.currAttributeCursor = None
        self.prefAttributeCursor = None
        self.init_default_attributes()
        self.prevLastCursorAttribute = None
        self.currLastCursorAttribute = None

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def set_last_cursor_attribute(self, lastCursorAttribute):
        self.prevLastCursorAttribute = self.currLastCursorAttribute
        self.currLastCursorAttribute = lastCursorAttribute

    def reset_last_cursor_attribute(self):
        self.prevLastCursorAttribute = None
        self.currLastCursorAttribute = None

    def is_last_cursor_attribute_change(self):
        if self.prevLastCursorAttribute is None:
            return False
        return self.prevLastCursorAttribute != self.currLastCursorAttribute

    def get_curr_attribute_cursor(self):
        return self.currAttributeCursor

    def is_attribute_cursor_active(self):
        return self.currAttributeCursor is not None

    def is_attribute_change(self):
        if not self.prevAttributes:
            return False
        return self.currAttributes != self.prevAttributes

    def reset_attribute_all(self):
        self.reset_attribute_delta()
        self.reset_attribute_cursor()

    def get_attribute_delta(self):
        return self.currAttributeDelta

    def reset_attribute_delta(self):
        self.currAttributeDelta = ""

    def set_attribute_delta(self, currAttributeDelta):
        self.currAttributeDelta = currAttributeDelta

    def reset_attribute_cursor(self):
        self.currAttributeCursor = None
        self.prefAttributeCursor = None

    def set_attribute_cursor(self, currAttributeCursor):
        self.prefAttributeCursor = self.currAttributeCursor
        self.currAttributeCursor = currAttributeCursor.copy()

    def reset_attributes(self, currAttributes):
        self.prevAttributes = None
        self.currAttributes = currAttributes

    def set_attributes(self, currAttributes):
        self.prevAttributes = self.currAttributes
        self.currAttributes = currAttributes.copy()

    def get_attribute_by_xy(self, x, y):
        if not self.currAttributes:
            return None
        if len(self.currAttributes) < y:
            return None
        if len(self.currAttributes[y]) < x - 1:
            return None
        try:
            return self.currAttributes[y][x]
        except KeyError:
            try:
                return self.defaultAttributes[0]
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "AttributeManager get_attribute_by_xy: Error accessing default attributes: "
                    + str(e),
                    debug.DebugLevel.ERROR,
                )
        return None

    def append_default_attributes(self, attribute):
        if not attribute:
            return
        if len(attribute) != 10:
            return
        self.defaultAttributes.append(attribute)

    def init_default_attributes(self):
        self.defaultAttributes = [None]
        self.defaultAttributes.append(
            [
                "default",  # fg
                "default",  # bg
                False,  # bold
                False,  # italics
                False,  # underscore
                False,  # strikethrough
                False,  # reverse
                False,  # blink
                "default",  # fontsize
                "default",  # fontfamily
            ]
        )  # end attribute

    def is_default_attribute(self, attribute):
        return attribute in self.defaultAttributes

    def has_attributes(self, cursor, update=True):
        if not cursor:
            return False
        cursor_pos = cursor.copy()
        try:
            attribute = self.get_attribute_by_xy(
                cursor_pos["x"], cursor_pos["y"]
            )

            if update:
                self.set_last_cursor_attribute(attribute)
            if not self.is_last_cursor_attribute_change():
                return False

            if self.is_default_attribute(attribute):
                return False

        except Exception as e:
            return False
        return True

    def format_attributes(self, attribute, attribute_format_string=""):
        # "black",
        # "red",
        # "green",
        # "brown",
        # "blue",
        # "magenta",
        # "cyan",
        # "white",
        # "default" # white.
        # _order_
        # "fg",
        # "bg",
        # "bold",
        # "italics",
        # "underscore",
        # "strikethrough",
        # "reverse",
        # "blink"
        # "fontsieze"
        # "fontfamily"
        if attribute_format_string == "":
            attribute_format_string = self.env["runtime"][
                "SettingsManager"
            ].get_setting("general", "attribute_format_string")
        if not attribute_format_string:
            return ""
        if attribute_format_string == "":
            return ""
        if not attribute:
            return ""
        if len(attribute) != 10:
            return ""

        # 0 FG color (name)
        try:
            attribute_format_string = attribute_format_string.replace(
                "fenrirFGColor", _(attribute[0])
            )
        except Exception as e:
            attribute_format_string = attribute_format_string.replace(
                "fenrirFGColor", ""
            )

        # 1 BG color (name)
        try:
            attribute_format_string = attribute_format_string.replace(
                "fenrirBGColor", _(attribute[1])
            )
        except Exception as e:
            attribute_format_string = attribute_format_string.replace(
                "fenrirBGColor", ""
            )

        # 2 bold (True/ False)
        try:
            if attribute[2]:
                attribute_format_string = attribute_format_string.replace(
                    "fenrirBold", _("bold")
                )
        except Exception as e:
            pass
        attribute_format_string = attribute_format_string.replace(
            "fenrirBold", ""
        )

        # 3 italics (True/ False)
        try:
            if attribute[3]:
                attribute_format_string = attribute_format_string.replace(
                    "fenrirItalics", _("italic")
                )
        except Exception as e:
            pass
        attribute_format_string = attribute_format_string.replace(
            "fenrirItalics", ""
        )

        # 4 underline (True/ False)
        try:
            if attribute[4]:
                attribute_format_string = attribute_format_string.replace(
                    "fenrirUnderline", _("underline")
                )
        except Exception as e:
            pass
        attribute_format_string = attribute_format_string.replace(
            "fenrirUnderline", ""
        )

        # 5 strikethrough (True/ False)
        try:
            if attribute[5]:
                attribute_format_string = attribute_format_string.replace(
                    "fenrirStrikethrough", _("strikethrough")
                )
        except Exception as e:
            pass
        attribute_format_string = attribute_format_string.replace(
            "fenrirStrikethrough", ""
        )

        # 6 reverse (True/ False)
        try:
            if attribute[6]:
                attribute_format_string = attribute_format_string.replace(
                    "fenrirReverse", _("reverse")
                )
        except Exception as e:
            pass
        attribute_format_string = attribute_format_string.replace(
            "fenrirReverse", ""
        )

        # 7 blink (True/ False)
        try:
            if attribute[7]:
                attribute_format_string = attribute_format_string.replace(
                    "fenrirBlink", _("blink")
                )
        except Exception as e:
            pass
        attribute_format_string = attribute_format_string.replace(
            "fenrirBlink", ""
        )

        # 8 font size (int/ string)
        try:
            try:
                attribute_format_string = attribute_format_string.replace(
                    "fenrirFontSize", int(attribute[8])
                )
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "AttributeManager formatAttributeToString: Error formatting font size as int: "
                    + str(e),
                    debug.DebugLevel.ERROR,
                )
            try:
                attribute_format_string = attribute_format_string.replace(
                    "fenrirFontSize", str(attribute[8])
                )
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "AttributeManager formatAttributeToString: Error formatting font size as string: "
                    + str(e),
                    debug.DebugLevel.ERROR,
                )
        except Exception as e:
            pass
        attribute_format_string = attribute_format_string.replace(
            "fenrirFontSize", _("default")
        )

        # 9 font family (string)
        try:
            attribute_format_string = attribute_format_string.replace(
                "fenrirFont", attribute[9]
            )
        except Exception as e:
            pass
        attribute_format_string = attribute_format_string.replace(
            "fenrirFont", _("default")
        )

        return attribute_format_string

    def track_highlights(self):
        """
        Detects text with changed attributes (highlighting) between screen updates.

        This is crucial for screen readers to announce when text becomes highlighted,
        selected, or changes visual emphasis (bold, reverse video, color changes, etc.)

        Enhanced version includes bracket pattern detection for better context in applications
        like ninjam that use patterns like [X]mute or [ ]mute.

        Returns:
            tuple: (highlighted_text, cursor_position)
                - highlighted_text: string of characters that gained highlighting
                - cursor_position: dict {'x': col, 'y': row} of first highlighted char
        """
        result = ""
        curr_cursor = None

        # Early exit conditions - no attribute comparison possible
        if self.prevAttributes is None:
            # First screen load - no previous attributes to compare against
            return result, curr_cursor
        if self.prevAttributes == self.currAttributes:
            # No attribute changes detected
            return result, curr_cursor
        if self.currAttributes is None:
            # Error condition - current attributes missing
            return result, curr_cursor
        if len(self.currAttributes) == 0:
            # Special case for PTY environments with no text content
            return result, curr_cursor

        # Get current screen text to correlate with attribute changes
        text = self.env["runtime"]["ScreenManager"].get_screen_text()
        text_lines = text.split("\n")

        # Sanity check: text lines must match attribute array dimensions
        if len(text_lines) != len(self.currAttributes):
            return result, curr_cursor

        # Track highlighted positions for context analysis
        highlighted_positions = []
        
        # Compare attributes line by line, character by character
        for line in range(len(self.prevAttributes)):
            if self.prevAttributes[line] != self.currAttributes[line]:
                # This line has attribute changes - examine each character
                # position
                for column in range(len(self.prevAttributes[line])):
                    if (
                        self.prevAttributes[line][column]
                        == self.currAttributes[line][column]
                    ):
                        # No change at this position
                        continue
                    # Attribute changed at this position - check if it's worth
                    # announcing
                    if self.is_useful_for_tracking(
                        line, column, self.currAttributes, self.prevAttributes
                    ):
                        # First highlighted character becomes cursor position
                        # for navigation
                        if not curr_cursor:
                            curr_cursor = {"x": column, "y": line}
                        # Store position for context analysis
                        highlighted_positions.append((line, column))
                        # Accumulate highlighted characters
                        result += text_lines[line][column]
                # Add space between lines of highlighted text for speech
                # clarity
                result += " "
        
        # Enhanced bracket pattern detection for better context
        if highlighted_positions:
            try:
                enhanced_result = self._detect_bracket_context(text_lines, highlighted_positions, result)
                if enhanced_result and enhanced_result != result:
                    # Debug logging for bracket detection
                    try:
                        self.env["runtime"]["DebugManager"].write_debug_out(
                            f"AttributeManager bracket detection: Original='{result}' Enhanced='{enhanced_result}'",
                            debug.DebugLevel.INFO
                        )
                    except Exception:
                        pass  # Don't let debug logging break functionality
                    result = enhanced_result
            except Exception as e:
                # If bracket detection fails, fall back to original result
                try:
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        f"AttributeManager bracket detection error: {e}",
                        debug.DebugLevel.ERROR
                    )
                except Exception:
                    pass  # Don't let debug logging break functionality
        
        return result, curr_cursor

    def _detect_bracket_context(self, text_lines, highlighted_positions, original_result):
        """
        Analyzes highlighted positions to detect bracket patterns and provide better context.
        
        This method specifically looks for patterns like [X]mute, [ ]mute, [X]xmit, etc.
        that are common in applications like ninjam where the bracket content indicates
        state but the meaningful context is the surrounding text.
        
        Args:
            text_lines: List of text lines from the screen
            highlighted_positions: List of (line, column) tuples of highlighted characters
            original_result: The original highlighted text result
            
        Returns:
            str: Enhanced result with context, or None if no bracket pattern detected
        """
        if not highlighted_positions:
            return None
            
        # Group positions by line for easier analysis
        positions_by_line = {}
        for line, column in highlighted_positions:
            if line not in positions_by_line:
                positions_by_line[line] = []
            positions_by_line[line].append(column)
        
        enhanced_results = []
        
        for line_num, columns in positions_by_line.items():
            if line_num >= len(text_lines):
                continue
                
            line_text = text_lines[line_num]
            columns.sort()  # Process columns in order
            
            # Look for bracket patterns in this line
            bracket_context = self._analyze_bracket_pattern(line_text, columns)
            if bracket_context:
                enhanced_results.append(bracket_context)
        
        # If we found enhanced context, return it; otherwise return None to use original
        if enhanced_results:
            # Remove duplicates while preserving order
            unique_results = []
            seen = set()
            for result in enhanced_results:
                if result not in seen:
                    unique_results.append(result)
                    seen.add(result)
            return " ".join(unique_results)
        return None
    
    def _analyze_bracket_pattern(self, line_text, highlighted_columns):
        """
        Analyzes a single line for bracket patterns around highlighted positions.
        
        Looks for patterns like:
        - master: [X]mute -> "master mute on"
        - metronome: [ ]mute -> "metronome mute off"  
        - [Left] [X]xmit -> "Left xmit on"
        
        Args:
            line_text: The text of the line
            highlighted_columns: List of column positions that are highlighted
            
        Returns:
            str: Context-aware description, or None if no pattern found
        """
        if not line_text or not highlighted_columns:
            return None
            
        # Look for bracket patterns around highlighted positions
        # Process columns in order and only return the first meaningful match
        for col in highlighted_columns:
            bracket_info = self._find_bracket_at_position(line_text, col)
            if bracket_info:
                bracket_start, bracket_end, bracket_content = bracket_info
                
                # Get context before and after the bracket
                context_before = self._get_context_before(line_text, bracket_start)
                context_after = self._get_context_after(line_text, bracket_end)
                
                # Build meaningful description
                description = self._build_bracket_description(
                    context_before, bracket_content, context_after
                )
                
                # Only return if we have meaningful context (not just state)
                if description and (context_before or context_after):
                    return description
        
        return None
    
    def _find_bracket_at_position(self, line_text, position):
        """
        Determines if a position is within a bracket pattern like [X] or [ ].
        
        Args:
            line_text: The text line
            position: Column position to check
            
        Returns:
            tuple: (bracket_start, bracket_end, bracket_content) or None
        """
        if position >= len(line_text):
            return None
            
        # Look for opening bracket before or at position
        bracket_start = None
        for i in range(max(0, position - 2), min(len(line_text), position + 3)):
            if line_text[i] == '[':
                bracket_start = i
                break
        
        if bracket_start is None:
            return None
            
        # Look for closing bracket after bracket_start
        bracket_end = None
        for i in range(bracket_start + 1, min(len(line_text), bracket_start + 5)):
            if line_text[i] == ']':
                bracket_end = i
                break
        
        if bracket_end is None:
            return None
            
        # Check if our position is within the bracket
        if bracket_start <= position <= bracket_end:
            bracket_content = line_text[bracket_start + 1:bracket_end]
            
            # Filter out brackets that are likely not controls
            # Skip brackets that contain complex content like "[-10.5dB center]"
            if len(bracket_content) > 3 and ('dB' in bracket_content or 'Hz' in bracket_content):
                return None
                
            return bracket_start, bracket_end, bracket_content
        
        return None
    
    def _get_context_before(self, line_text, bracket_start):
        """
        Gets meaningful context words before a bracket.
        
        Args:
            line_text: The text line
            bracket_start: Position of opening bracket
            
        Returns:
            str: Context words before bracket, or empty string
        """
        if bracket_start <= 0:
            return ""
            
        # Get text before bracket
        before_text = line_text[:bracket_start].rstrip()
        
        # Special handling for common ninjam patterns
        # Look for patterns like "master: " or "metronome: " or bracketed labels
        
        # Check if we have a colon-separated label immediately before
        if ':' in before_text:
            # Get the last colon-separated part
            parts = before_text.split('|')  # Split by pipe first
            last_part = parts[-1].strip()
            
            if ':' in last_part:
                # Extract the label before the colon
                label_parts = last_part.split(':')
                if len(label_parts) >= 2:
                    return label_parts[-2].strip()
        
        # Look for bracketed content that might be a label
        # Pattern: [Something] [X]target -> "Something"
        bracket_matches = []
        i = 0
        while i < len(before_text):
            if before_text[i] == '[':
                start = i
                i += 1
                while i < len(before_text) and before_text[i] != ']':
                    i += 1
                if i < len(before_text):  # Found closing bracket
                    content = before_text[start+1:i]
                    if content.strip():
                        bracket_matches.append(content.strip())
            i += 1
        
        # If we found bracketed content, use the last one
        if bracket_matches:
            return bracket_matches[-1]
        
        # Fall back to removing bracketed content and getting words
        cleaned_text = ""
        bracket_level = 0
        for char in before_text:
            if char == '[':
                bracket_level += 1
            elif char == ']':
                bracket_level -= 1
            elif bracket_level == 0:
                cleaned_text += char
        
        # Clean up separators and get meaningful words
        words = []
        for word in cleaned_text.replace(':', '').replace('|', '').strip().split():
            if word and not word.startswith('[') and not word.endswith(']'):
                words.append(word)
        
        # Return last word as context
        if words:
            return words[-1]
        
        return ""
    
    def _get_context_after(self, line_text, bracket_end):
        """
        Gets meaningful context words after a bracket.
        
        Args:
            line_text: The text line  
            bracket_end: Position of closing bracket
            
        Returns:
            str: Context words after bracket, or empty string
        """
        if bracket_end >= len(line_text) - 1:
            return ""
            
        # Get text after bracket and find first meaningful word
        after_text = line_text[bracket_end + 1:].lstrip()
        
        # Get first word, removing common separators
        words = after_text.replace(':', '').replace('|', '').strip().split()
        
        if words:
            return words[0]
        
        return ""
    
    def _build_bracket_description(self, context_before, bracket_content, context_after):
        """
        Builds a human-readable description from bracket context.
        
        Args:
            context_before: Words before the bracket
            bracket_content: Content inside brackets (X, space, etc.)
            context_after: Words after the bracket
            
        Returns:
            str: Human-readable description
        """
        # Clean up bracket content
        bracket_content = bracket_content.strip()
        
        # Determine state based on bracket content
        if bracket_content == 'X':
            state = "on"
        elif bracket_content == '' or bracket_content == ' ':
            state = "off"
        else:
            state = bracket_content  # For other patterns
        
        # Build description prioritizing context_before, then context_after
        # Add spaces between components for better speech clarity
        components = []
        
        if context_before:
            components.append(context_before)
        if context_after:
            components.append(context_after)
        components.append(state)
        
        # Join with spaces for better speech flow
        return " ".join(components)

    def is_useful_for_tracking(
        self,
        line,
        column,
        currAttributes,
        prevAttributes,
        attribute=1,
        mode="zaxe",
    ):
        """
        Determines if an attribute change at a specific position is worth announcing.

        This prevents announcing every minor attribute change and focuses on meaningful
        highlighting that indicates user selections, focus changes, or important emphasis.

        Args:
            line, column: Position of the attribute change
            currAttributes, prevAttributes: Current and previous attribute arrays
            attribute: Which attribute to examine (1=background color by default)
            mode: Detection algorithm ('zaxe', 'default', 'barrier')

        Returns:
            bool: True if this attribute change should be announced to user
        """
        # Sanity checks for valid position and sufficient screen content
        if len(currAttributes) <= 3:
            return False
        if line < 0:
            return False
        if line > len(currAttributes):
            return False

        useful = False

        if mode == "default":
            # Simple mode: announce any non-default attribute
            useful = not self.is_default_attribute(
                currAttributes[line][column]
            )

        elif (mode == "zaxe") or (mode == ""):
            # Context-aware mode: only announce attributes that stand out from surroundings
            # This prevents announcing entire blocks of highlighted text character by character
            # by checking if the attribute differs from adjacent lines

            if line == 0:
                # Top line: compare against lines below
                useful = (
                    currAttributes[line][column][attribute]
                    != currAttributes[line + 1][column][attribute]
                ) and (
                    currAttributes[line][column][attribute]
                    != currAttributes[line + 2][column][attribute]
                )
            elif line >= len(prevAttributes):
                # Bottom line: compare against lines above
                useful = (
                    currAttributes[line][column][attribute]
                    != currAttributes[line - 1][column][attribute]
                ) and (
                    currAttributes[line][column][attribute]
                    != currAttributes[line - 2][column][attribute]
                )
            else:
                # Middle lines: compare against both directions
                useful = (
                    currAttributes[line][column][attribute]
                    != currAttributes[line + 1][column][attribute]
                ) and (
                    currAttributes[line][column][attribute]
                    != currAttributes[line - 1][column][attribute]
                )

        elif mode == "barrier":
            # Barrier mode: future enhancement for detecting screen
            # boundaries/separators
            useful = True

        return useful



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.


from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return "No Description found"

    def run(self):
        if not self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "keyboard", "charDeleteEcho"
        ):
            return
        # detect typing or chilling
        if (
            self.env["screen"]["new_cursor"]["x"]
            >= self.env["screen"]["old_cursor"]["x"]
        ):
            return

        # More than just a deletion happend
        if self.env["runtime"]["ScreenManager"].is_delta(ignoreSpace=True):
            return

        # no deletion
        if not self.env["runtime"]["ScreenManager"].is_negative_delta():
            return

        # too much for a single backspace...
        # word begin produce a diff wiht len == 2 |a | others with 1 |a|
        if len(self.env["screen"]["newNegativeDelta"]) > 2:
            return

        curr_negative_delta = self.env["screen"]["newNegativeDelta"]
        if (
            len(curr_negative_delta.strip()) != len(curr_negative_delta)
            and curr_negative_delta.strip() != ""
        ):
            curr_negative_delta = curr_negative_delta.strip()
        self.env["runtime"]["OutputManager"].present_text(
            curr_negative_delta,
            interrupt=True,
            ignore_punctuation=True,
            announce_capital=True,
            flush=False,
        )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core.i18n import _
from fenrirscreenreader.utils import char_utils


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _("set review and phonetically presents the current character")

    def run(self):
        self.env["runtime"][
            "CursorManager"
        ].enter_review_mode_curr_text_cursor()

        (
            self.env["screen"]["newCursorReview"]["x"],
            self.env["screen"]["newCursorReview"]["y"],
            curr_char,
        ) = char_utils.get_current_char(
            self.env["screen"]["newCursorReview"]["x"],
            self.env["screen"]["newCursorReview"]["y"],
            self.env["screen"]["new_content_text"],
        )

        if curr_char.isspace():
            self.env["runtime"]["OutputManager"].present_text(
                _("blank"), interrupt=True, flush=False
            )
        else:
            curr_char = char_utils.get_phonetic(curr_char)
            self.env["runtime"]["OutputManager"].present_text(
                curr_char, interrupt=True, announce_capital=True, flush=False
            )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import os
import os.path
import select
import socket

from fenrirscreenreader.core import debug
from fenrirscreenreader.core.eventData import FenrirEventType
from fenrirscreenreader.core.remoteDriver import RemoteDriver as remoteDriver


class driver(remoteDriver):
    def __init__(self):
        remoteDriver.__init__(self)

    def initialize(self, environment):
        self.env = environment
        # Use threading instead of multiprocessing to avoid pickle issues
        # with self.env (which contains unpicklable file handles)
        self.env["runtime"]["ProcessManager"].add_custom_event_thread(
            self.watch_dog, multiprocess=False
        )

    def watch_dog(self, active, event_queue):
        # echo "command say this is a test" | nc localhost 22447
        self.fenrirSock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.fenrirSock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.fenrirSock.settimeout(5.0)  # Prevent hanging on slow clients
        self.host = "127.0.0.1"
        self.port = self.env["runtime"]["SettingsManager"].get_setting_as_int(
            "remote", "port"
        )
        self.fenrirSock.bind((self.host, self.port))
        self.fenrirSock.listen(1)
        while active.value:
            try:
                r, _, _ = select.select([self.fenrirSock], [], [], 0.8)
            except select.error:
                break
            if r == []:
                continue
            if self.fenrirSock in r:
                client_sock, client_addr = self.fenrirSock.accept()
                # Ensure client socket is always closed to prevent resource
                # leaks
                try:
                    try:
                        rawdata = client_sock.recv(8129)
                    except Exception as e:
                        self.env["runtime"]["DebugManager"].write_debug_out(
                            "tcpDriver watch_dog: Error receiving data from "
                            "client: "
                            + str(e),
                            debug.DebugLevel.ERROR,
                        )
                        rawdata = b""  # Set default empty data if recv fails

                    try:
                        data = rawdata.decode("utf-8").rstrip().lstrip()
                        event_queue.put(
                            {"Type": FenrirEventType.remote_incomming, "data": data}
                        )
                    except Exception as e:
                        self.env["runtime"]["DebugManager"].write_debug_out(
                            "tcpDriver watch_dog: Error decoding/queuing data: "
                            + str(e),
                            debug.DebugLevel.ERROR,
                        )
                finally:
                    # Always close client socket, even if data processing fails
                    try:
                        client_sock.close()
                    except Exception as e:
                        self.env["runtime"]["DebugManager"].write_debug_out(
                            "tcpDriver watch_dog: Error closing client socket: "
                            + str(e),
                            debug.DebugLevel.ERROR,
                        )
        if self.fenrirSock:
            self.fenrirSock.close()
            self.fenrirSock = None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import time

from fenrirscreenreader.core import debug
from fenrirscreenreader.core.inputDriver import InputDriver as inputDriver


class driver(inputDriver):
    def __init__(self):
        inputDriver.__init__(self)

    def initialize(self, environment):
        self.env = environment
        self.env["runtime"]["InputManager"].set_shortcut_type("KEY")
        self._initialized = True
        print("Input Debug Driver: Initialized")

    def shutdown(self):
        if self._initialized:
            self.remove_all_devices()
        self._initialized = False
        print("Input Debug Driver: Shutdown")

    def get_input_event(self):
        time.sleep(0.1)
        if not self._initialized:
            return None
        print("Input Debug Driver: get_input_event")
        return None

    def write_event_buffer(self):
        if not self._initialized:
            return
        print("Input Debug Driver: write_event_buffer")

    def clear_event_buffer(self):
        if not self._initialized:
            return
        del self.env["input"]["eventBuffer"][:]
        print("Input Debug Driver: clear_event_buffer")

    def update_input_devices(self, new_devices=None, init=False):
        if not self._initialized:
            return
        print("Input Debug Driver: update_input_devices")

    def get_led_state(self, led=0):
        if not self._initialized:
            return False
        return False

    def toggle_led_state(self, led=0):
        if not self._initialized:
            return
        print("Input Debug Driver: toggle_led_state")

    def grab_all_devices(self):
        if not self._initialized:
            return
        print("Input Debug Driver: grab_all_devices")

    def ungrab_all_devices(self):
        if not self._initialized:
            return
        print("Input Debug Driver: ungrab_all_devices")

    def remove_all_devices(self):
        if not self._initialized:
            return
        print("Input Debug Driver: remove_all_devices")

    def __del__(self):
        if self._initialized:
            self.remove_all_devices()
        print("Input Debug Driver: __del__")



#!/usr/bin/env python3

from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return "Apply the last tested voice from safe voice browser"

    def run(self):
        try:
            # Check if we have a tested voice
            if (
                "commandBuffer" not in self.env
                or "lastTestedModule" not in self.env["commandBuffer"]
                or "lastTestedVoice" not in self.env["commandBuffer"]
            ):
                self.env["runtime"]["OutputManager"].present_text(
                    "No voice has been tested yet", interrupt=True
                )
                self.env["runtime"]["OutputManager"].present_text(
                    "Use voice browser first", interrupt=True
                )
                return

            module = self.env["commandBuffer"]["lastTestedModule"]
            voice = self.env["commandBuffer"]["lastTestedVoice"]

            self.env["runtime"]["OutputManager"].present_text(
                f"Applying {voice} from {module}", interrupt=True
            )

            # Apply to runtime settings only (temporary until saved)
            SettingsManager = self.env["runtime"]["SettingsManager"]

            # Store old values for safety
            old_driver = SettingsManager.get_setting("speech", "driver")
            old_module = SettingsManager.get_setting("speech", "module")
            old_voice = SettingsManager.get_setting("speech", "voice")

            try:
                # Apply new settings to runtime only (use set_setting to update
                # settingArgDict)
                SettingsManager.set_setting(
                    "speech", "driver", "speechdDriver"
                )
                SettingsManager.set_setting("speech", "module", module)
                SettingsManager.set_setting("speech", "voice", voice)

                # Apply to speech driver instance directly
                if "SpeechDriver" in self.env["runtime"]:
                    SpeechDriver = self.env["runtime"]["SpeechDriver"]

                    # Set the module and voice on the driver instance
                    SpeechDriver.set_module(module)
                    SpeechDriver.set_voice(voice)

                    self.env["runtime"]["OutputManager"].present_text(
                        "Voice applied successfully!", interrupt=True
                    )
                    self.env["runtime"]["OutputManager"].present_text(
                        "Use save settings to make permanent", interrupt=True
                    )
                    self.env["runtime"]["OutputManager"].play_sound("Accept")

            except Exception as e:
                # Revert on failure
                SettingsManager.set_setting("speech", "driver", old_driver)
                SettingsManager.set_setting("speech", "module", old_module)
                SettingsManager.set_setting("speech", "voice", old_voice)

                self.env["runtime"]["OutputManager"].present_text(
                    f"Failed to apply voice, reverted: {str(e)}",
                    interrupt=True,
                )
                self.env["runtime"]["OutputManager"].play_sound("Error")

        except Exception as e:
            self.env["runtime"]["OutputManager"].present_text(
                f"Apply voice error: {str(e)}", interrupt=True
            )
            self.env["runtime"]["OutputManager"].play_sound("Error")

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# Debugger module for the Fenrir screen reader.

import os
import pathlib
from datetime import datetime

from fenrirscreenreader.core import debug


class DebugManager:
    def __init__(self, file_name=""):
        self._file = None
        self._fileOpened = False
        self._fileName = (
            "/tmp/fenrir_"
            + str(os.getpid())
            + "_"
            + str(datetime.utcnow().strftime("%Y-%m-%d_%H-%M-%S"))
            + ".log"
        )
        if file_name != "":
            self._fileName = file_name

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        self.close_debug_file()

    def __del__(self):
        try:
            self.shutdown()
        except Exception as e:
            # Cannot use debug manager to log its own shutdown errors
            print(f"DebugManager shutdown: Error during shutdown: {e}")

    def open_debug_file(self, file_name=""):
        self._fileOpened = False
        if file_name != "":
            self._fileName = file_name
        if self._fileName != "":
            directory = os.path.dirname(self._fileName)
            if not os.path.exists(directory):
                pathlib.Path(directory).mkdir(parents=True, exist_ok=True)
            try:
                self._file = open(self._fileName, "a")
                # Set readable permissions for all users (644)
                os.chmod(self._fileName, 0o644)
                self._fileOpened = True
            except Exception as e:
                print(e)

    def write_debug_out(
        self, text, level=debug.DebugLevel.DEACTIVE, on_any_level=False
    ):

        mode = self.env["runtime"]["SettingsManager"].get_setting(
            "general", "debug_mode"
        )
        if mode == "":
            mode = "FILE"
        mode = mode.upper().split(",")
        file_mode = "FILE" in mode
        print_mode = "PRINT" in mode

        if (
            int(
                self.env["runtime"]["SettingsManager"].get_setting_as_int(
                    "general", "debug_level"
                )
            )
            < int(level)
        ) and not (
            on_any_level
            and int(
                self.env["runtime"]["SettingsManager"].get_setting_as_int(
                    "general", "debug_level"
                )
            )
            > int(debug.DebugLevel.DEACTIVE)
        ):
            if self._fileOpened:
                self.close_debug_file()
            return
        else:
            if not self._fileOpened and file_mode:
                self.open_debug_file()
            timestamp = str(datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S.%f"))
            if on_any_level:
                level_info = "INFO ANY"
            else:
                level_info = str(level)

            # Changed order: text comes first, then level and timestamp
            msg = text + " - " + level_info + " " + timestamp

            if print_mode:
                print(msg)
            if file_mode:
                try:
                    self._file.write(msg + "\n")
                except Exception as e:
                    print(e)

    def close_debug_file(self):
        if not self._fileOpened:
            return False
        if self._file is not None:
            try:
                self._file.close()
            except Exception as e:
                print(e)
        self._fileOpened = False
        return True

    def get_debug_file(self):
        return self._fileName

    def set_debug_file(self, file_name):
        self.close_debug_file()
        self._fileName = file_name



"""
Integration tests for remote control functionality.

Tests the remote control system including Unix socket and TCP communication,
command parsing, and settings management.
"""

import pytest
import socket
import time
from unittest.mock import Mock, patch

from fenrirscreenreader.core.remoteManager import RemoteManager


@pytest.mark.integration
@pytest.mark.remote
class TestRemoteCommandParsing:
    """Test remote control command parsing and execution."""

    def setup_method(self):
        """Create RemoteManager instance for each test."""
        self.manager = RemoteManager()

    def test_say_command_parsing(self, mock_environment):
        """Test parsing of 'command say' messages."""
        self.manager.initialize(mock_environment)

        result = self.manager.handle_command_execution_with_response(
            "say Hello World"
        )

        assert result["success"] is True
        assert "Speaking" in result["message"]
        mock_environment["runtime"]["OutputManager"].speak_text.assert_called_once_with(
            "Hello World"
        )

    def test_interrupt_command(self, mock_environment):
        """Test speech interruption command."""
        self.manager.initialize(mock_environment)

        result = self.manager.handle_command_execution_with_response("interrupt")

        assert result["success"] is True
        mock_environment["runtime"][
            "OutputManager"
        ].interrupt_output.assert_called_once()

    def test_setting_change_parsing(self, mock_environment):
        """Test parsing of 'setting set' commands."""
        self.manager.initialize(mock_environment)

        # Mock parse_setting_args to verify it gets called
        with patch.object(
            mock_environment["runtime"]["SettingsManager"], "parse_setting_args"
        ) as mock_parse:
            result = self.manager.handle_settings_change_with_response(
                "set speech#rate=0.8"
            )

            assert result["success"] is True
            mock_parse.assert_called_once_with("speech#rate=0.8")

    def test_clipboard_command(self, mock_environment):
        """Test clipboard setting via remote control."""
        self.manager.initialize(mock_environment)

        result = self.manager.handle_command_execution_with_response(
            "clipboard Test clipboard content"
        )

        assert result["success"] is True
        mock_environment["runtime"][
            "MemoryManager"
        ].add_value_to_first_index.assert_called_once_with(
            "clipboardHistory", "Test clipboard content"
        )

    def test_quit_command(self, mock_environment):
        """Test Fenrir shutdown command."""
        self.manager.initialize(mock_environment)

        result = self.manager.handle_command_execution_with_response(
            "quitapplication"
        )

        assert result["success"] is True
        mock_environment["runtime"][
            "EventManager"
        ].stop_main_event_loop.assert_called_once()

    def test_unknown_command_rejection(self, mock_environment):
        """Test that unknown commands are rejected."""
        self.manager.initialize(mock_environment)

        result = self.manager.handle_command_execution_with_response(
            "unknown_command"
        )

        assert result["success"] is False
        assert "Unknown command" in result["message"]


@pytest.mark.integration
@pytest.mark.remote
class TestRemoteSettingsControl:
    """Test remote control of settings."""

    def setup_method(self):
        """Create RemoteManager instance for each test."""
        self.manager = RemoteManager()

    def test_setting_reset(self, mock_environment):
        """Test resetting settings to defaults."""
        self.manager.initialize(mock_environment)

        with patch.object(
            mock_environment["runtime"]["SettingsManager"], "reset_setting_arg_dict"
        ) as mock_reset:
            result = self.manager.handle_settings_change_with_response("reset")

            assert result["success"] is True
            mock_reset.assert_called_once()

    def test_setting_save(self, mock_environment):
        """Test saving settings to file."""
        self.manager.initialize(mock_environment)

        mock_environment["runtime"]["SettingsManager"].get_settings_file = Mock(
            return_value="/tmp/test.conf"
        )

        with patch.object(
            mock_environment["runtime"]["SettingsManager"], "save_settings"
        ) as mock_save:
            result = self.manager.handle_settings_change_with_response("save")

            assert result["success"] is True
            mock_save.assert_called_once()

    def test_settings_remote_disabled(self, mock_environment):
        """Test that settings commands are blocked when disabled."""
        mock_environment["runtime"]["SettingsManager"].get_setting_as_bool = Mock(
            return_value=False
        )
        self.manager.initialize(mock_environment)

        result = self.manager.handle_settings_change_with_response(
            "set speech#rate=0.5"
        )

        assert result["success"] is False
        assert "disabled" in result["message"].lower()


@pytest.mark.integration
@pytest.mark.remote
class TestRemoteDataFormat:
    """Test remote control data format handling."""

    def setup_method(self):
        """Create RemoteManager instance for each test."""
        self.manager = RemoteManager()

    def test_command_prefix_case_insensitive(self, mock_environment):
        """Test that command prefixes are case-insensitive."""
        self.manager.initialize(mock_environment)

        # All of these should work
        result1 = self.manager.handle_remote_incomming_with_response(
            "COMMAND say test"
        )
        result2 = self.manager.handle_remote_incomming_with_response(
            "command say test"
        )
        result3 = self.manager.handle_remote_incomming_with_response(
            "CoMmAnD say test"
        )

        assert all(r["success"] for r in [result1, result2, result3])

    def test_setting_prefix_case_insensitive(self, mock_environment):
        """Test that setting prefixes are case-insensitive."""
        self.manager.initialize(mock_environment)

        with patch.object(
            mock_environment["runtime"]["SettingsManager"], "parse_setting_args"
        ):
            result1 = self.manager.handle_remote_incomming_with_response(
                "SETTING set speech#rate=0.5"
            )
            result2 = self.manager.handle_remote_incomming_with_response(
                "setting set speech#rate=0.5"
            )

            assert all(r["success"] for r in [result1, result2])

    def test_empty_data_handling(self, mock_environment):
        """Test handling of empty remote data."""
        self.manager.initialize(mock_environment)

        result = self.manager.handle_remote_incomming_with_response("")

        assert result["success"] is False
        assert "No data" in result["message"]

    def test_invalid_format_rejection(self, mock_environment):
        """Test rejection of invalid command format."""
        self.manager.initialize(mock_environment)

        result = self.manager.handle_remote_incomming_with_response(
            "invalid format without prefix"
        )

        assert result["success"] is False
        assert "Unknown command format" in result["message"]


@pytest.mark.integration
@pytest.mark.remote
class TestWindowDefinition:
    """Test window definition via remote control."""

    def setup_method(self):
        """Create RemoteManager instance for each test."""
        self.manager = RemoteManager()

    def test_define_window_valid_coordinates(self, mock_environment):
        """Test defining a window with valid coordinates."""
        self.manager.initialize(mock_environment)

        result = self.manager.handle_command_execution_with_response(
            "window 10 5 70 20"
        )

        assert result["success"] is True
        mock_environment["runtime"][
            "CursorManager"
        ].set_window_for_application.assert_called_once()

        # Verify the coordinates were parsed correctly
        call_args = mock_environment["runtime"][
            "CursorManager"
        ].set_window_for_application.call_args
        start, end = call_args[0]
        assert start == {"x": 10, "y": 5}
        assert end == {"x": 70, "y": 20}

    def test_define_window_insufficient_coordinates(self, mock_environment):
        """Test that window definition with < 4 coordinates is ignored."""
        self.manager.initialize(mock_environment)

        # Should succeed but not call set_window_for_application
        result = self.manager.handle_command_execution_with_response("window 10 20")

        assert result["success"] is True
        mock_environment["runtime"][
            "CursorManager"
        ].set_window_for_application.assert_not_called()

    def test_reset_window(self, mock_environment):
        """Test resetting window to full screen."""
        self.manager.initialize(mock_environment)

        result = self.manager.handle_command_execution_with_response("resetwindow")

        assert result["success"] is True
        mock_environment["runtime"][
            "CursorManager"
        ].clear_window_for_application.assert_called_once()


@pytest.mark.integration
@pytest.mark.remote
class TestVMenuControl:
    """Test VMenu control via remote."""

    def setup_method(self):
        """Create RemoteManager instance for each test."""
        self.manager = RemoteManager()

    def test_set_vmenu(self, mock_environment):
        """Test setting VMenu to specific menu."""
        self.manager.initialize(mock_environment)

        result = self.manager.handle_command_execution_with_response(
            "vmenu /vim/file/save"
        )

        assert result["success"] is True
        mock_environment["runtime"]["VmenuManager"].set_curr_menu.assert_called_once_with(
            "/vim/file/save"
        )

    def test_reset_vmenu(self, mock_environment):
        """Test resetting VMenu to default."""
        self.manager.initialize(mock_environment)

        result = self.manager.handle_command_execution_with_response("resetvmenu")

        assert result["success"] is True
        mock_environment["runtime"]["VmenuManager"].set_curr_menu.assert_called_once_with()


@pytest.mark.integration
@pytest.mark.remote
@pytest.mark.slow
class TestRemoteControlThroughput:
    """Test remote control performance characteristics."""

    def test_rapid_say_commands(self, mock_environment):
        """Test handling of rapid successive say commands."""
        manager = RemoteManager()
        manager.initialize(mock_environment)

        # Send 100 rapid commands
        for i in range(100):
            result = manager.handle_command_execution_with_response(f"say test {i}")
            assert result["success"] is True

        # Verify all were queued
        assert (
            mock_environment["runtime"]["OutputManager"].speak_text.call_count == 100
        )

    def test_rapid_setting_changes(self, mock_environment):
        """Test handling of rapid setting changes."""
        manager = RemoteManager()
        manager.initialize(mock_environment)

        # Rapidly change speech rate
        with patch.object(
            mock_environment["runtime"]["SettingsManager"], "parse_setting_args"
        ) as mock_parse:
            for rate in [0.5, 0.6, 0.7, 0.8, 0.9, 1.0]:
                result = manager.handle_settings_change_with_response(
                    f"set speech#rate={rate}"
                )
                assert result["success"] is True

            assert mock_parse.call_count == 6



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.


from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _("get next quick menu entry")

    def run(self):
        menu = ""
        value = ""
        if self.env["runtime"]["QuickMenuManager"].next_entry():
            menu = self.env["runtime"]["QuickMenuManager"].get_current_entry()
            if menu != "":
                value = self.env["runtime"][
                    "QuickMenuManager"
                ].get_current_value()
                self.env["runtime"]["OutputManager"].present_text(
                    menu + " " + value, interrupt=True
                )
        else:
            self.env["runtime"]["OutputManager"].present_text(
                _("Quick menu not available"), interrupt=True
            )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3

import subprocess
import time

from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment
        self.testMessage = "This is a voice test. The quick brown fox jumps over the lazy dog."
        self.modules = []
        self.voices = []
        self.module_index = 0
        self.voice_index = 0
        self.browserActive = False
        self.originalBindings = {}
        self.lastAnnounceTime = 0

    def shutdown(self):
        pass

    def get_description(self):
        return "Interactive voice browser with arrow key navigation"

    def run(self):
        if self.browserActive:
            self.exit_voice_browser()
            return

        self.env["runtime"]["OutputManager"].present_text(
            "Starting voice browser", interrupt=True
        )

        # Load modules
        self.modules = self.get_speechd_modules()
        if not self.modules:
            self.env["runtime"]["OutputManager"].present_text(
                "No speech modules found", interrupt=True
            )
            self.env["runtime"]["OutputManager"].play_sound("Error")
            return

        # Set current module
        current_module = self.env["runtime"]["SettingsManager"].get_setting(
            "speech", "module"
        )
        if current_module and current_module in self.modules:
            self.module_index = self.modules.index(current_module)

        # Load voices
        self.load_voices_for_current_module()

        # Set current voice
        current_voice = self.env["runtime"]["SettingsManager"].get_setting(
            "speech", "voice"
        )
        if current_voice and current_voice in self.voices:
            self.voice_index = self.voices.index(current_voice)

        # Enter browser mode
        self.enter_voice_browser()
        self.announce_current_selection()

    def enter_voice_browser(self):
        """Enter voice browser mode with custom key bindings"""
        self.browserActive = True

        # Store original bindings
        self.originalBindings = self.env["bindings"].copy()

        # Override navigation keys for voice browsing
        # Use lambda to capture self and create bound methods
        self.env["bindings"][str([1, ["KEY_UP"]])] = "VOICE_BROWSER_PREV_VOICE"
        self.env["bindings"][
            str([1, ["KEY_DOWN"]])
        ] = "VOICE_BROWSER_NEXT_VOICE"
        self.env["bindings"][
            str([1, ["KEY_LEFT"]])
        ] = "VOICE_BROWSER_PREV_MODULE"
        self.env["bindings"][
            str([1, ["KEY_RIGHT"]])
        ] = "VOICE_BROWSER_NEXT_MODULE"
        self.env["bindings"][str([1, ["KEY_ENTER"]])] = "VOICE_BROWSER_TEST"
        self.env["bindings"][str([1, ["KEY_SPACE"]])] = "VOICE_BROWSER_APPLY"
        self.env["bindings"][str([1, ["KEY_ESC"]])] = "VOICE_BROWSER_EXIT"

        # Register browser commands
        self.register_browser_commands()

        self.env["runtime"]["OutputManager"].present_text(
            "Voice browser active", interrupt=True
        )
        self.env["runtime"]["OutputManager"].present_text(
            "Up/Down=voices, Left/Right=modules, Enter=test, Space=apply, Esc=exit",
            interrupt=True,
        )

    def register_browser_commands(self):
        """Register voice browser commands with the command manager"""
        # Create command objects for voice browser actions
        CommandManager = self.env["runtime"]["CommandManager"]

        # This is a hack - we'll store references to our methods in the environment
        # so the key handlers can call them
        if "voiceBrowserInstance" not in self.env["runtime"]:
            self.env["runtime"]["voiceBrowserInstance"] = self

    def exit_voice_browser(self):
        """Exit voice browser and restore normal key bindings"""
        if not self.browserActive:
            return

        self.browserActive = False

        # Restore original bindings
        self.env["bindings"] = self.originalBindings

        # Clean up
        if "voiceBrowserInstance" in self.env["runtime"]:
            del self.env["runtime"]["voiceBrowserInstance"]

        self.env["runtime"]["OutputManager"].present_text(
            "Voice browser exited", interrupt=True
        )

    def load_voices_for_current_module(self):
        """Load voices for current module"""
        if self.module_index < len(self.modules):
            module = self.modules[self.module_index]
            self.voices = self.get_module_voices(module)
            self.voice_index = 0  # Reset to first voice when changing modules

    def announce_current_selection(self):
        """Announce current module and voice"""
        # Throttle announcements to avoid spam
        now = time.time()
        if now - self.lastAnnounceTime < 0.3:
            return
        self.lastAnnounceTime = now

        if not self.modules or self.module_index >= len(self.modules):
            return

        module = self.modules[self.module_index]
        if self.voices and self.voice_index < len(self.voices):
            voice = self.voices[self.voice_index]
            self.env["runtime"]["OutputManager"].present_text(
                f"{module}: {voice} ({self.voice_index + 1}/{len(self.voices)})",
                interrupt=True,
            )
        else:
            self.env["runtime"]["OutputManager"].present_text(
                f"{module}: No voices", interrupt=True
            )

    def next_voice(self):
        """Move to next voice"""
        if not self.voices:
            return
        self.voice_index = (self.voice_index + 1) % len(self.voices)
        self.announce_current_selection()

    def prev_voice(self):
        """Move to previous voice"""
        if not self.voices:
            return
        self.voice_index = (self.voice_index - 1) % len(self.voices)
        self.announce_current_selection()

    def next_module(self):
        """Move to next module"""
        self.module_index = (self.module_index + 1) % len(self.modules)
        self.load_voices_for_current_module()
        self.announce_current_selection()

    def prev_module(self):
        """Move to previous module"""
        self.module_index = (self.module_index - 1) % len(self.modules)
        self.load_voices_for_current_module()
        self.announce_current_selection()

    def test_voice(self):
        """Test current voice"""
        if not self.voices or self.voice_index >= len(self.voices):
            self.env["runtime"]["OutputManager"].present_text(
                "No voice selected", interrupt=True
            )
            return

        module = self.modules[self.module_index]
        voice = self.voices[self.voice_index]

        self.env["runtime"]["OutputManager"].present_text(
            "Testing...", interrupt=True
        )
        if self.preview_voice(module, voice):
            # Store for apply command
            self.env["commandBuffer"]["lastTestedModule"] = module
            self.env["commandBuffer"]["lastTestedVoice"] = voice
            self.env["runtime"]["OutputManager"].play_sound("Accept")
        else:
            self.env["runtime"]["OutputManager"].play_sound("Error")

    def apply_voice(self):
        """Apply current voice to Fenrir"""
        if not self.voices or self.voice_index >= len(self.voices):
            return

        module = self.modules[self.module_index]
        voice = self.voices[self.voice_index]

        try:
            SettingsManager = self.env["runtime"]["SettingsManager"]
            SettingsManager.settings["speech"]["driver"] = "speechdDriver"
            SettingsManager.settings["speech"]["module"] = module
            SettingsManager.settings["speech"]["voice"] = voice

            if "SpeechDriver" in self.env["runtime"]:
                SpeechDriver = self.env["runtime"]["SpeechDriver"]
                SpeechDriver.shutdown()
                SpeechDriver.initialize(self.env)

                self.env["runtime"]["OutputManager"].present_text(
                    "Voice applied!", interrupt=True
                )
                self.env["runtime"]["OutputManager"].play_sound("Accept")

        except Exception as e:
            self.env["runtime"]["OutputManager"].present_text(
                "Apply failed", interrupt=True
            )
            self.env["runtime"]["OutputManager"].play_sound("Error")

    def preview_voice(self, module, voice):
        """Test voice with spd-say"""
        try:
            cmd = ["spd-say", "-o", module, "-y", voice, self.testMessage]
            result = subprocess.run(cmd, timeout=10)
            return result.returncode == 0
        except Exception:
            return False

    def get_speechd_modules(self):
        """Get available speech modules"""
        try:
            result = subprocess.run(
                ["spd-say", "-O"], capture_output=True, text=True, timeout=10
            )
            if result.returncode == 0:
                lines = result.stdout.strip().split("\n")
                return [line.strip() for line in lines[1:] if line.strip()]
        except Exception:
            pass
        return []

    def get_module_voices(self, module):
        """Get voices for module"""
        try:
            result = subprocess.run(
                ["spd-say", "-o", module, "-L"],
                capture_output=True,
                text=True,
                timeout=10,
            )
            if result.returncode == 0:
                lines = result.stdout.strip().split("\n")
                voices = []
                for line in lines[1:]:
                    if not line.strip():
                        continue
                    if module.lower() == "espeak-ng":
                        voice = self.process_espeak_voice(line)
                        if voice:
                            voices.append(voice)
                    else:
                        voices.append(line.strip())
                return voices
        except Exception:
            pass
        return []

    def process_espeak_voice(self, voiceLine):
        """Process espeak voice format"""
        parts = [p for p in voiceLine.split() if p]
        if len(parts) < 2:
            return None
        lang_code = parts[-2].lower()
        variant = parts[-1].lower()
        return (
            f"{lang_code}+{variant}"
            if variant and variant != "none"
            else lang_code
        )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import importlib.util
import os

from fenrirscreenreader.core.i18n import _

_base_path = os.path.join(os.path.dirname(__file__), "bookmark_base.py")
_spec = importlib.util.spec_from_file_location("bookmark_base", _base_path)
_module = importlib.util.module_from_spec(_spec)
_spec.loader.exec_module(_module)
bookmark_command = _module.bookmark_command


class command(bookmark_command):
    def __init__(self):
        super().__init__(7, "read")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import getpass
import os
import string
import time
from select import select

from fenrirscreenreader.core import debug


def remove_nonprintable(text):
    # Get the difference of all ASCII characters from the set of printable
    # characters
    nonprintable = set([chr(i) for i in range(128)]).difference(
        string.printable
    )
    # Use translate to remove all non-printable characters
    return text.translate({ord(character): None for character in nonprintable})


def insert_newlines(string, every=64):
    return "\n".join(
        string[i : i + every] for i in range(0, len(string), every)
    )


def split_every(toSplit, every=64):
    return list(toSplit[i : i + every] for i in range(0, len(toSplit), every))


def create_screen_event_data(content):
    event_data = {
        "bytes": content,
        "lines": content["lines"],
        "columns": content["columns"],
        "textCursor": {
            "x": int(content["cursor"][0]),
            "y": int(content["cursor"][1]),
        },
        "screen": content["screen"],
        "text": content["text"],
        "attributes": content["attributes"],
        "screenUpdateTime": time.time(),
    }
    return event_data.copy()


def has_more(fd, timetout=0.05):
    r, _, _ = select([fd], [], [], timetout)
    return fd in r


def has_more_what(fd_list, timetout=0.05):
    if not isinstance(fd_list, list):
        return []
    elif fd_list == []:
        return []
    r, _, _ = select(fd_list, [], [], timetout)
    return r


def is_valid_shell(shell=""):
    if not isinstance(shell, str):
        return False
    if shell == "":
        return False
    try:
        if not os.path.isfile(shell):
            return False
        if not os.access(shell, os.X_OK):
            return False
    except Exception as e:
        # Utility function, no env access - use fallback logging
        print(
            f"screen_utils is_valid_shell: Error checking shell {shell}: {e}"
        )
        return False
    return True


def get_shell():
    try:
        shell = os.environ["FENRIRSHELL"]
        if is_valid_shell(shell):
            return shell
    except Exception as e:
        # Utility function, no env access - continue silently
        pass
    try:
        shell = os.environ["SHELL"]
        if is_valid_shell(shell):
            return shell
    except Exception as e:
        # Utility function, no env access - continue silently
        pass
    try:
        if os.access("/etc/passwd", os.R_OK):
            with open("/etc/passwd") as f:
                users = f.readlines()
                for user in users:
                    (username, encrypwd, uid, gid, gecos, homedir, shell) = (
                        user.split(":")
                    )
                    shell = shell.replace("\n", "")
                    if username == getpass.getuser():
                        if is_valid_shell(shell):
                            return shell
    except Exception as e:
        # Utility function, no env access - continue silently
        pass
    if is_valid_shell("/bin/bash"):
        return "/bin/bash"
    return "/bin/sh"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core.i18n import _
from fenrirscreenreader.utils import char_utils
from fenrirscreenreader.utils import line_utils


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _("Move Review to the first character on the line")

    def run(self):
        # Check if we're in table mode first
        is_table_mode = self.env["runtime"]["TableManager"].is_table_mode()
        if is_table_mode:
            table_info = self.env["runtime"]["TableManager"].move_to_first_char_in_cell()
            if table_info:
                char_utils.present_char_for_review(
                    self.env,
                    table_info['character'],
                    interrupt=True,
                    announce_capital=True,
                    flush=False,
                )
                self.env["runtime"]["OutputManager"].present_text(
                    _("first character in cell {0}").format(table_info['column_header']),
                    interrupt=False,
                )
            else:
                self.env["runtime"]["OutputManager"].present_text(
                    _("no table data"), interrupt=True, flush=False
                )
            return
        
        # Regular line first character navigation (when not in table mode)
        cursor_pos = self.env["runtime"][
            "CursorManager"
        ].get_review_or_text_cursor()
        x, y, curr_line = line_utils.get_current_line(
            cursor_pos["x"],
            cursor_pos["y"],
            self.env["screen"]["new_content_text"],
        )
        if curr_line.isspace():
            self.env["runtime"]["OutputManager"].present_text(
                _("line is empty"), interrupt=True
            )
            return
        self.env["runtime"]["CursorManager"].set_review_cursor_position(
            (len(curr_line) - len(curr_line.lstrip())), cursor_pos["y"]
        )
        (
            self.env["screen"]["newCursorReview"]["x"],
            self.env["screen"]["newCursorReview"]["y"],
            curr_char,
        ) = char_utils.get_current_char(
            self.env["screen"]["newCursorReview"]["x"],
            self.env["screen"]["newCursorReview"]["y"],
            self.env["screen"]["new_content_text"],
        )

        char_utils.present_char_for_review(
            self.env,
            curr_char,
            interrupt=True,
            announce_capital=True,
            flush=False,
        )
        self.env["runtime"]["OutputManager"].present_text(
            _("first character in line indent {0}").format(
                str(len(curr_line) - len(curr_line.lstrip()))
            ),
            interrupt=False,
        )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import _thread
import importlib
import os

import pyperclip

from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment, script_path=""):
        self.env = environment
        self.script_path = script_path

    def shutdown(self):
        pass

    def get_description(self):
        return _("Export current fenrir clipboard to X or GUI clipboard")

    def run(self):
        _thread.start_new_thread(self._thread_run, ())

    def _thread_run(self):
        try:
            # Check if clipboard is empty
            if self.env["runtime"]["MemoryManager"].is_index_list_empty(
                "clipboardHistory"
            ):
                self.env["runtime"]["OutputManager"].present_text(
                    _("clipboard empty"), interrupt=True
                )
                return

            # Get current clipboard content
            clipboard = self.env["runtime"][
                "MemoryManager"
            ].get_index_list_element("clipboardHistory")

            # Remember original display environment variable if it exists
            original_display = os.environ.get("DISPLAY", "")
            success = False

            # Try different display options
            for i in range(10):
                display = f":{i}"
                try:
                    # Set display environment variable
                    os.environ["DISPLAY"] = display
                    # Attempt to set clipboard content
                    # Weird workaround for some distros
                    importlib.reload(pyperclip)
                    pyperclip.copy(clipboard)
                    # If we get here without exception, we found a working
                    # display
                    success = True
                    break
                except Exception:
                    # Failed for this display, try next one
                    continue

            # Restore original display setting
            if original_display:
                os.environ["DISPLAY"] = original_display
            else:
                os.environ.pop("DISPLAY", None)

            # Notify the user of the result
            if success:
                self.env["runtime"]["OutputManager"].present_text(
                    _("exported to the X session."), interrupt=True
                )
            else:
                self.env["runtime"]["OutputManager"].present_text(
                    _(
                        "failed to export to X clipboard. No available display "
                        "found."
                    ),
                    interrupt=True,
                )

        except Exception as e:
            self.env["runtime"]["OutputManager"].present_text(
                str(e), sound_icon="", interrupt=False
            )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
"""
Fenrir Release Validation Tool

Comprehensive validation suite for Fenrir releases, including syntax validation,
dependency checking, import testing, and basic functionality validation.

Usage:
    python3 tools/validate_release.py                    # Full validation
    python3 tools/validate_release.py --quick           # Skip slow tests
    python3 tools/validate_release.py --fix             # Auto-fix issues where possible
"""

import ast
import os
import sys
import argparse
import subprocess
import tempfile
import importlib.util
from pathlib import Path
import time


class ReleaseValidator:
    def __init__(self, verbose=True):
        self.verbose = verbose
        self.errors = []
        self.warnings = []
        self.fixes_applied = []
        self.tests_run = 0
        self.tests_passed = 0
        
    def log(self, message, level="INFO"):
        """Log a message with appropriate formatting."""
        if not self.verbose and level == "INFO":
            return
            
        colors = {
            "INFO": "\033[0;36m",     # Cyan
            "SUCCESS": "\033[0;32m",  # Green
            "WARNING": "\033[1;33m",  # Yellow
            "ERROR": "\033[0;31m",    # Red
            "HEADER": "\033[1;34m",   # Bold Blue
        }
        
        reset = "\033[0m"
        color = colors.get(level, "")
        
        if level == "HEADER":
            print(f"\n{color}{'='*60}")
            print(f"{message}")
            print(f"{'='*60}{reset}")
        else:
            symbol = {
                "SUCCESS": "✓",
                "ERROR": "✗", 
                "WARNING": "⚠",
                "INFO": "•"
            }.get(level, "•")
            
            print(f"{color}{symbol} {message}{reset}")
    
    def run_test(self, name, test_func, *args, **kwargs):
        """Run a test and track results."""
        self.tests_run += 1
        try:
            result = test_func(*args, **kwargs)
            if result:
                self.tests_passed += 1
                self.log(f"{name}: PASSED", "SUCCESS")
            else:
                self.log(f"{name}: FAILED", "ERROR")
            return result
        except Exception as e:
            self.log(f"{name}: ERROR - {e}", "ERROR")
            self.errors.append(f"{name}: {e}")
            return False
    
    def validate_python_syntax(self, directory, fix_mode=False):
        """Validate Python syntax across all files."""
        python_files = list(Path(directory).rglob("*.py"))
        # Filter out cache and build directories
        python_files = [f for f in python_files if not any(part.startswith(('__pycache__', '.git', 'build', 'dist')) for part in f.parts)]
        
        syntax_errors = []
        fixed_files = []
        
        for filepath in python_files:
            try:
                with open(filepath, 'r', encoding='utf-8') as f:
                    content = f.read()
                ast.parse(content, filename=str(filepath))
            except SyntaxError as e:
                syntax_errors.append((filepath, e))
                
                if fix_mode:
                    # Try to fix common f-string issues
                    fixed_content = self.fix_fstring_issues(content)
                    if fixed_content != content:
                        try:
                            ast.parse(fixed_content, filename=str(filepath))
                            # Fix worked, write it back
                            with open(filepath, 'w', encoding='utf-8') as f:
                                f.write(fixed_content)
                            fixed_files.append(filepath)
                            syntax_errors.pop()  # Remove from errors
                        except SyntaxError:
                            pass  # Fix didn't work
            except Exception as e:
                syntax_errors.append((filepath, e))
        
        if fixed_files:
            self.fixes_applied.extend([f"Fixed f-string syntax in {f}" for f in fixed_files])
        
        if syntax_errors:
            for filepath, error in syntax_errors[:5]:  # Show only first 5
                if isinstance(error, SyntaxError):
                    self.errors.append(f"Syntax error in {filepath}:{error.lineno}: {error.msg}")
                else:
                    self.errors.append(f"Error in {filepath}: {error}")
            if len(syntax_errors) > 5:
                self.errors.append(f"... and {len(syntax_errors) - 5} more syntax errors")
        
        return len(syntax_errors) == 0
    
    def fix_fstring_issues(self, content):
        """Fix common f-string syntax issues."""
        lines = content.split('\n')
        
        for i, line in enumerate(lines):
            # Look for f-strings that span multiple lines incorrectly
            if ('f"' in line and line.count('"') % 2 == 1 and 
                i + 1 < len(lines) and lines[i + 1].strip()):
                
                next_line = lines[i + 1]
                # Common patterns to fix
                if (next_line.strip().endswith('}"') or 
                    'str(e)}' in next_line or
                    next_line.strip().startswith(('fenrirVersion.', 'self.'))):
                    
                    # Merge the lines properly
                    fixed_line = line.rstrip() + next_line.strip()
                    lines[i] = fixed_line
                    lines[i + 1] = ''
        
        return '\n'.join(line for line in lines if line.strip() or not line)
    
    def validate_dependencies(self):
        """Run the existing dependency checker."""
        try:
            result = subprocess.run([
                sys.executable, "check-dependencies.py"
            ], capture_output=True, text=True, timeout=30)
            
            if result.returncode == 0:
                return True
            else:
                self.errors.append(f"Dependency check failed: {result.stderr}")
                return False
        except subprocess.TimeoutExpired:
            self.errors.append("Dependency check timed out")
            return False
        except Exception as e:
            self.errors.append(f"Could not run dependency check: {e}")
            return False
    
    def validate_core_imports(self):
        """Test importing core Fenrir modules."""
        # Change to src directory for imports
        original_path = sys.path.copy()
        src_dir = Path.cwd() / "src"
        if src_dir.exists():
            sys.path.insert(0, str(src_dir))
        
        core_modules = [
            "fenrirscreenreader.core.fenrirManager",
            "fenrirscreenreader.core.commandManager",
            "fenrirscreenreader.core.eventManager",
            "fenrirscreenreader.core.screenManager",
            "fenrirscreenreader.core.inputManager",
            "fenrirscreenreader.core.outputManager",
        ]
        
        import_failures = []
        
        for module_name in core_modules:
            try:
                importlib.import_module(module_name)
            except ImportError as e:
                import_failures.append(f"{module_name}: {e}")
            except Exception as e:
                import_failures.append(f"{module_name}: Unexpected error: {e}")
        
        # Restore path
        sys.path = original_path
        
        if import_failures:
            self.errors.extend(import_failures)
            return False
        return True
    
    def validate_command_structure(self):
        """Validate command file structure and naming."""
        commands_dir = Path("src/fenrirscreenreader/commands")
        if not commands_dir.exists():
            self.errors.append("Commands directory not found")
            return False
        
        issues = []
        
        # Check command directories
        expected_dirs = ["commands", "onHeartBeat", "onKeyInput", "onCursorChange", 
                        "onScreenUpdate", "onScreenChanged", "vmenu-profiles"]
        
        for expected_dir in expected_dirs:
            if not (commands_dir / expected_dir).exists():
                issues.append(f"Missing expected directory: {expected_dir}")
        
        # Check for critical issues only (skip template files and base classes)
        for py_file in commands_dir.rglob("*.py"):
            if (py_file.name.startswith("__") or 
                "template" in py_file.name.lower() or
                "base" in py_file.name.lower()):
                continue
                
            try:
                with open(py_file, 'r', encoding='utf-8') as f:
                    content = f.read()
                
                # Critical structure checks only
                if "class command" not in content:
                    issues.append(f"{py_file}: Missing 'class command' definition")
                
                # Skip method checks for files that inherit from base classes
                if ("super().__init__" in content or 
                    "importlib.util" in content or
                    "_base.py" in content):
                    continue  # These inherit methods from base classes
                
                # Only check direct implementations
                # Special case: Application profile commands use load/unload instead of run
                if "onSwitchApplicationProfile" in str(py_file):
                    if "def load" not in content and "def unload" not in content:
                        issues.append(f"{py_file}: Missing 'load' or 'unload' method")
                else:
                    critical_methods = ["run"]  # Focus on the most critical method
                    for method in critical_methods:
                        if (f"def {method}" not in content and 
                            "super()" not in content):  # Skip if uses inheritance
                            issues.append(f"{py_file}: Missing '{method}' method")
                        
            except Exception as e:
                issues.append(f"{py_file}: Could not validate structure: {e}")
        
        # Only report critical issues, not template/base class warnings
        critical_issues = [issue for issue in issues if not any(skip in issue.lower() 
                          for skip in ["template", "base", "missing 'initialize'", "missing 'shutdown'"])]
        
        if critical_issues:
            self.warnings.extend(critical_issues[:5])  # Limit warnings
            if len(critical_issues) > 5:
                self.warnings.append(f"... and {len(critical_issues) - 5} more critical command structure issues")
        
        # Return success if no critical issues (warnings are acceptable)
        return len(critical_issues) == 0
    
    def validate_configuration_files(self):
        """Validate configuration file structure."""
        config_dir = Path("config")
        if not config_dir.exists():
            self.errors.append("Config directory not found")
            return False
        
        required_configs = [
            "settings/settings.conf",
            "keyboard/desktop.conf", 
            "punctuation/default.conf"
        ]
        
        missing_configs = []
        for config_file in required_configs:
            if not (config_dir / config_file).exists():
                missing_configs.append(config_file)
        
        if missing_configs:
            self.errors.extend([f"Missing config file: {f}" for f in missing_configs])
            return False
        
        return True
    
    def validate_installation_scripts(self):
        """Validate installation and setup scripts."""
        required_scripts = ["setup.py", "install.sh", "uninstall.sh"]
        missing_scripts = []
        
        for script in required_scripts:
            if not Path(script).exists():
                missing_scripts.append(script)
        
        if missing_scripts:
            self.warnings.extend([f"Missing installation script: {s}" for s in missing_scripts])
        
        # Check setup.py syntax if it exists
        if Path("setup.py").exists():
            try:
                with open("setup.py", 'r') as f:
                    content = f.read()
                ast.parse(content, filename="setup.py")
            except SyntaxError as e:
                self.errors.append(f"setup.py syntax error: {e}")
                return False
        
        return len(missing_scripts) == 0
    
    def validate_repository_cleanliness(self):
        """Check for cache files and other artifacts that shouldn't be in git."""
        # Check for Python cache files in git tracking
        try:
            result = subprocess.run([
                "git", "ls-files", "--cached"
            ], capture_output=True, text=True, timeout=10)
            
            if result.returncode == 0:
                tracked_files = result.stdout.strip().split('\n')
                cache_files = [f for f in tracked_files if '__pycache__' in f or f.endswith('.pyc')]
                
                if cache_files:
                    self.errors.extend([f"Python cache file tracked in git: {f}" for f in cache_files[:5]])
                    if len(cache_files) > 5:
                        self.errors.append(f"... and {len(cache_files) - 5} more cache files in git")
                    return False
                else:
                    return True
            else:
                self.warnings.append("Could not check git tracked files")
                return True
                
        except subprocess.TimeoutExpired:
            self.warnings.append("Git check timed out")
            return True
        except Exception as e:
            self.warnings.append(f"Could not check repository cleanliness: {e}")
            return True
    
    def generate_report(self):
        """Generate final validation report."""
        self.log("FENRIR RELEASE VALIDATION REPORT", "HEADER")
        
        # Test Summary
        success_rate = (self.tests_passed / self.tests_run * 100) if self.tests_run > 0 else 0
        self.log(f"Tests run: {self.tests_run}")
        self.log(f"Tests passed: {self.tests_passed}")
        self.log(f"Success rate: {success_rate:.1f}%")
        
        # Fixes Applied
        if self.fixes_applied:
            self.log("\nAUTO-FIXES APPLIED:", "HEADER")
            for fix in self.fixes_applied:
                self.log(fix, "SUCCESS")
        
        # Errors
        if self.errors:
            self.log(f"\nERRORS ({len(self.errors)}):", "HEADER")
            for error in self.errors:
                self.log(error, "ERROR")
        
        # Warnings  
        if self.warnings:
            self.log(f"\nWARNINGS ({len(self.warnings)}):", "HEADER")
            for warning in self.warnings:
                self.log(warning, "WARNING")
        
        # Final Status
        if not self.errors and success_rate >= 80:
            self.log("\n🎉 RELEASE VALIDATION PASSED", "SUCCESS")
            self.log("The codebase appears ready for release", "SUCCESS")
            return True
        elif not self.errors:
            self.log("\n⚠️ RELEASE VALIDATION PASSED WITH WARNINGS", "WARNING")
            self.log("Release is possible but issues should be addressed", "WARNING")
            return True
        else:
            self.log("\n❌ RELEASE VALIDATION FAILED", "ERROR")
            self.log("Critical issues must be fixed before release", "ERROR")
            return False


def main():
    parser = argparse.ArgumentParser(description='Comprehensive Fenrir release validation')
    parser.add_argument('--quick', action='store_true',
                       help='Skip slow tests (dependency checking)')
    parser.add_argument('--fix', action='store_true',
                       help='Attempt to fix issues automatically where possible')
    parser.add_argument('--quiet', action='store_true',
                       help='Reduce output verbosity')
    
    args = parser.parse_args()
    
    # Ensure we're in the project root
    if not Path("src/fenrirscreenreader").exists():
        print("Error: Must be run from Fenrir project root directory")
        sys.exit(1)
    
    validator = ReleaseValidator(verbose=not args.quiet)
    
    validator.log("FENRIR RELEASE VALIDATION STARTING", "HEADER")
    start_time = time.time()
    
    # Run validation tests
    validator.run_test(
        "Python syntax validation",
        validator.validate_python_syntax,
        "src/fenrirscreenreader",
        args.fix
    )
    
    if not args.quick:
        validator.run_test(
            "Dependency validation", 
            validator.validate_dependencies
        )
    
    validator.run_test(
        "Core module imports",
        validator.validate_core_imports
    )
    
    validator.run_test(
        "Command structure validation",
        validator.validate_command_structure
    )
    
    validator.run_test(
        "Configuration files validation",
        validator.validate_configuration_files
    )
    
    validator.run_test(
        "Installation scripts validation",
        validator.validate_installation_scripts
    )
    
    validator.run_test(
        "Repository cleanliness validation",
        validator.validate_repository_cleanliness
    )
    
    # Generate final report
    elapsed_time = time.time() - start_time
    validator.log(f"\nValidation completed in {elapsed_time:.1f} seconds")
    
    success = validator.generate_report()
    
    sys.exit(0 if success else 1)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import builtins
import gettext
import locale

# the only function we really need to call here is gettext.install.
# Python3 has simplified this.
gettext.install("fenrir")

# Make _ available for import to satisfy static analysis tools
_ = builtins._



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import time
from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _("Handles delayed retry for tab completion detection")

    def run(self):
        """Check for and process pending tab completions with slight delay"""
        # Only process if we have tab completion state
        if "tabCompletion" not in self.env["commandBuffer"]:
            return

        tab_state = self.env["commandBuffer"]["tabCompletion"]
        pending = tab_state.get("pendingCompletion")
        
        if not pending:
            return

        current_time = time.time()
        
        # Process pending completion after 50ms delay
        if current_time - pending["timestamp"] < 0.05:
            return

        # Check if screen delta is now available
        if not self.env["runtime"]["ScreenManager"].is_delta():
            # Give up after 200ms total
            if current_time - pending["timestamp"] > 0.2:
                tab_state["pendingCompletion"] = None
                tab_state["retryCount"] = 0
            return

        # Process the delayed completion
        delta_text = self.env["screen"]["new_delta"]
        x_move = pending["x_move"]
        
        # Use the same flexible matching logic as main tab completion
        match_found = self._is_flexible_completion_match(x_move, delta_text)
        
        if not match_found:
            # Try pattern-based detection as final fallback
            match_found = self._detect_completion_patterns(delta_text)

        if match_found and delta_text:
            # Mark that we've handled this delta to prevent duplicate announcements
            tab_state["lastProcessedDelta"] = delta_text
            tab_state["lastProcessedTime"] = current_time
            
            # Filter and announce the completion
            curr_delta = delta_text
            if (len(curr_delta.strip()) != len(curr_delta) and 
                curr_delta.strip() != ""):
                curr_delta = curr_delta.strip()
            
            if curr_delta:
                self.env["runtime"]["OutputManager"].present_text(
                    curr_delta, interrupt=True, announce_capital=True, flush=False
                )

        # Clear pending completion
        tab_state["pendingCompletion"] = None
        tab_state["retryCount"] = 0

    def _is_flexible_completion_match(self, x_move, delta_text):
        """Use flexible matching (duplicated from main command for heartbeat use)"""
        if not delta_text:
            return False
        
        delta_len = len(delta_text)
        
        # Exact match
        if x_move == delta_len:
            return True
        
        # Flexible range: allow ±2 characters difference
        if abs(x_move - delta_len) <= 2 and delta_len > 0:
            return True
        
        # For longer completions, allow proportional variance
        if delta_len > 10 and abs(x_move - delta_len) <= (delta_len * 0.2):
            return True
        
        return False

    def _detect_completion_patterns(self, delta_text):
        """Detect common tab completion patterns (duplicated from main command)"""
        if not delta_text:
            return False
        
        delta_stripped = delta_text.strip()
        
        # File extension completion
        if '.' in delta_stripped and delta_stripped.count('.') <= 2:
            return True
        
        # Path completion
        if '/' in delta_stripped or '\\' in delta_stripped:
            return True
        
        # Command parameter completion
        if delta_stripped.startswith('-') and len(delta_stripped) > 1:
            return True
        
        # Word boundary completion
        if delta_stripped.isalnum() and len(delta_stripped) >= 2:
            return True
        
        return False

    def set_callback(self, callback):
        pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core.i18n import _
from fenrirscreenreader.utils import char_utils


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _("moves review to the previous character")

    def run(self):
        self.env["screen"]["oldCursorReview"] = self.env["screen"][
            "newCursorReview"
        ]
        if not self.env["screen"]["newCursorReview"]:
            self.env["screen"]["newCursorReview"] = self.env["screen"][
                "new_cursor"
            ].copy()

        # Check if we're in table mode for bounded navigation
        if self.env["runtime"]["TableManager"].is_table_mode():
            table_info = self.env["runtime"]["TableManager"].get_current_table_cell_info()
            if table_info:
                cursor_pos = self.env["screen"]["newCursorReview"]
                line_text = self.env["runtime"]["ScreenManager"].get_line_text(
                    cursor_pos["y"]
                )
                if line_text:
                    column_start = self.env["runtime"]["TableManager"].get_column_start_position(
                        line_text, table_info["column_index"]
                    )
                    
                    # Check if we're already at the start of the cell
                    if cursor_pos["x"] <= column_start:
                        # At cell boundary - announce start and don't move
                        char_utils.present_char_for_review(
                            self.env,
                            (
                                table_info["cell_content"][0] 
                                if table_info["cell_content"] else ""
                            ),
                            interrupt=True,
                            announce_capital=True,
                            flush=False,
                        )
                        self.env["runtime"]["OutputManager"].present_text(
                            _("start of cell"), 
                            interrupt=False, 
                            sound_icon="StartOfLine"
                        )
                        return
                    
                    # Move within cell bounds
                    cell_content = table_info["cell_content"]
                    relative_pos = cursor_pos["x"] - column_start
                    if relative_pos > 0:
                        new_relative_pos = relative_pos - 1
                        self.env["screen"]["newCursorReview"]["x"] = (
                            column_start + new_relative_pos
                        )
                        
                        # Get character at new position
                        if new_relative_pos < len(cell_content):
                            prev_char = cell_content[new_relative_pos]
                        else:
                            prev_char = ""
                        
                        char_utils.present_char_for_review(
                            self.env,
                            prev_char,
                            interrupt=True,
                            announce_capital=True,
                            flush=False,
                        )
                        return

        # Regular navigation for non-table mode
        (
            self.env["screen"]["newCursorReview"]["x"],
            self.env["screen"]["newCursorReview"]["y"],
            prev_char,
            end_of_screen,
            line_break,
        ) = char_utils.get_prev_char(
            self.env["screen"]["newCursorReview"]["x"],
            self.env["screen"]["newCursorReview"]["y"],
            self.env["screen"]["new_content_text"],
        )

        char_utils.present_char_for_review(
            self.env,
            prev_char,
            interrupt=True,
            announce_capital=True,
            flush=False,
        )
        if end_of_screen:
            if self.env["runtime"]["SettingsManager"].get_setting_as_bool(
                "review", "end_of_screen"
            ):
                self.env["runtime"]["OutputManager"].present_text(
                    _("end of screen"),
                    interrupt=True,
                    sound_icon="EndOfScreen",
                )
        if line_break:
            if self.env["runtime"]["SettingsManager"].get_setting_as_bool(
                "review", "line_break"
            ):
                self.env["runtime"]["OutputManager"].present_text(
                    _("line break"), interrupt=False, sound_icon="EndOfLine"
                )
        # is has attribute it enabled?
        if self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "general", "hasattributes"
        ):
            cursor_pos = self.env["screen"]["newCursorReview"]

            if not self.env["runtime"]["AttributeManager"].has_attributes(
                cursor_pos
            ):
                return
            self.env["runtime"]["OutputManager"].present_text(
                "has attribute", sound_icon="HasAttributes", interrupt=False
            )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core.i18n import _
from fenrirscreenreader.utils import mark_utils


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _("reads from the cursor to the bottom of the screen")

    def run(self):
        # Prefer review cursor over text cursor
        cursor_pos = self.env["runtime"][
            "CursorManager"
        ].get_review_or_text_cursor()

        text_after_cursor = mark_utils.get_text_after_mark(
            cursor_pos, self.env["screen"]["new_content_text"]
        )

        if text_after_cursor.isspace():
            self.env["runtime"]["OutputManager"].present_text(
                _("blank"), sound_icon="EmptyLine", interrupt=True
            )
        else:
            self.env["runtime"]["OutputManager"].present_text(
                text_after_cursor, interrupt=True
            )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3

import glob
import os

from fenrirscreenreader.core import debug


class DynamicKeyboardLayoutCommand:
    """Dynamic command class for keyboard layout selection"""

    def __init__(self, layoutName, layoutPath, env):
        self.layoutName = layoutName
        self.layoutPath = layoutPath
        # Extract just the base name without extension for comparison
        self.layoutBaseName = layoutName
        self.env = env

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return f"Set keyboard layout to {self.layoutName}"

    def run(self):
        try:
            settingsManager = self.env["runtime"]["SettingsManager"]
            currentLayout = settingsManager.get_setting(
                "keyboard", "keyboardLayout"
            )

            # Check if already set (compare both full path and base name)
            currentBaseName = os.path.splitext(os.path.basename(currentLayout))[0] if currentLayout else ""
            if currentBaseName.lower() == self.layoutBaseName.lower() or currentLayout.lower() == self.layoutPath.lower():
                self.env["runtime"]["OutputManager"].present_text(
                    f"Keyboard layout already set to {self.layoutName}"
                )
                return

            # Set the new layout in the config file using full path
            try:
                # Update the setting in memory
                settingsManager.set_setting(
                    "keyboard", "keyboardLayout", self.layoutPath
                )

                # Save to the actual config file
                configFilePath = settingsManager.get_settings_file()
                settingsManager.save_settings(configFilePath)

                self.env["runtime"]["OutputManager"].present_text(
                    f"Keyboard layout set to {self.layoutName}. Please restart Fenrir for this change to take effect."
                )
                # Play accept sound
                self.env["runtime"]["OutputManager"].present_text(
                    "", sound_icon="Accept", interrupt=False
                )
            except Exception as e:
                self.env["runtime"]["OutputManager"].present_text(
                    f"Failed to change keyboard layout to {self.layoutName}"
                )
                self.env["runtime"]["DebugManager"].write_debug_out(
                    f"DynamicKeyboardLayout: Error setting layout {self.layoutName}: {e}",
                    debug.DebugLevel.ERROR,
                )
                # Play error sound
                self.env["runtime"]["OutputManager"].present_text(
                    "", sound_icon="ErrorSound", interrupt=False
                )

        except Exception as e:
            self.env["runtime"]["OutputManager"].present_text(
                f"Keyboard layout change error: {str(e)}", interrupt=True
            )
            self.env["runtime"]["DebugManager"].write_debug_out(
                f"DynamicKeyboardLayout: Unexpected error for {self.layoutName}: {e}",
                debug.DebugLevel.ERROR,
            )

    def set_callback(self, callback):
        pass


def add_dynamic_keyboard_layout_menus(VmenuManager):
    """Add dynamic keyboard layout menus to vmenu system"""
    try:
        env = VmenuManager.env

        # Get keyboard layout files
        layouts = get_keyboard_layouts(env)
        if not layouts:
            return

        # Create keyboard layouts submenu
        layoutMenu = {}

        # Add layout commands
        for layoutName, layoutPath in layouts:
            layoutCommand = DynamicKeyboardLayoutCommand(
                layoutName, layoutPath, env
            )
            layoutMenu[f"{layoutName} Action"] = layoutCommand

        # Find keyboard menu in existing vmenu structure
        # If fenrir menu exists, add layouts under it
        if "fenrir Menu" in VmenuManager.menuDict:
            fenrirMenu = VmenuManager.menuDict["fenrir Menu"]
            if "keyboard Menu" in fenrirMenu:
                # Add dynamic layouts to existing keyboard menu
                keyboardMenu = fenrirMenu["keyboard Menu"]
                keyboardMenu["Keyboard Layouts Menu"] = layoutMenu
            else:
                # Create keyboard menu with layouts
                fenrirMenu["keyboard Menu"] = {
                    "Keyboard Layouts Menu": layoutMenu
                }
        else:
            # Create standalone keyboard layouts menu
            VmenuManager.menuDict["Keyboard Layouts Menu"] = layoutMenu

    except Exception as e:
        # Use debug manager for error logging
        if "DebugManager" in env.get("runtime", {}):
            env["runtime"]["DebugManager"].write_debug_out(
                f"Error creating dynamic keyboard layout menus: {e}",
                debug.DebugLevel.ERROR,
            )
        else:
            print(f"Error creating dynamic keyboard layout menus: {e}")


def get_keyboard_layouts(env):
    """Get available keyboard layouts from keyboard directory"""
    layouts = []

    try:
        # Get keyboard directory paths
        keyboardDirs = []

        # Check system installation path
        systemKeyboardPath = "/etc/fenrirscreenreader/keyboard/"
        if os.path.exists(systemKeyboardPath):
            keyboardDirs.append(systemKeyboardPath)

        # Check source/development path
        try:
            import fenrirscreenreader

            fenrirPath = os.path.dirname(fenrirscreenreader.__file__)
            devKeyboardPath = os.path.join(
                fenrirPath, "..", "..", "config", "keyboard"
            )
            devKeyboardPath = os.path.abspath(devKeyboardPath)
            if os.path.exists(devKeyboardPath):
                keyboardDirs.append(devKeyboardPath)
        except Exception:
            pass

        # Get current layout setting path
        try:
            currentLayoutSetting = env["runtime"]["SettingsManager"].get_setting(
                "keyboard", "keyboardLayout"
            )
            if currentLayoutSetting and os.path.exists(currentLayoutSetting):
                currentLayoutDir = os.path.dirname(currentLayoutSetting)
                if currentLayoutDir not in keyboardDirs:
                    keyboardDirs.append(currentLayoutDir)
        except Exception:
            pass

        # Scan for .conf files
        seenLayouts = set()
        for keyboardDir in keyboardDirs:
            try:
                confFiles = glob.glob(os.path.join(keyboardDir, "*.conf"))
                for confFile in confFiles:
                    layoutName = os.path.splitext(os.path.basename(confFile))[
                        0
                    ]
                    if layoutName not in seenLayouts:
                        seenLayouts.add(layoutName)
                        layouts.append((layoutName, confFile))
            except Exception as e:
                if "DebugManager" in env.get("runtime", {}):
                    env["runtime"]["DebugManager"].write_debug_out(
                        f"Error scanning keyboard directory {keyboardDir}: {e}",
                        debug.DebugLevel.WARNING,
                    )

        # Sort layouts alphabetically
        layouts.sort(key=lambda x: x[0].lower())

    except Exception as e:
        if "DebugManager" in env.get("runtime", {}):
            env["runtime"]["DebugManager"].write_debug_out(
                f"Error getting keyboard layouts: {e}",
                debug.DebugLevel.ERROR,
            )

    return layouts



"""
Shared pytest fixtures for Fenrir tests.

This file contains fixtures and configuration used across all test modules.
"""

import os
import sys
import tempfile
from pathlib import Path
from unittest.mock import MagicMock, Mock

import pytest

# Add src directory to Python path for imports
fenrir_root = Path(__file__).parent.parent
sys.path.insert(0, str(fenrir_root / "src"))


@pytest.fixture
def mock_environment():
    """Create a minimal mock environment for testing.

    Returns a mock environment dict with required runtime managers mocked.
    This allows testing components without initializing the full Fenrir stack.
    """
    env = {
        "runtime": {
            "DebugManager": Mock(write_debug_out=Mock()),
            "OutputManager": Mock(
                present_text=Mock(),
                speak_text=Mock(),
                interrupt_output=Mock(),
            ),
            "SettingsManager": Mock(
                get_setting=Mock(return_value="default"),
                get_setting_as_int=Mock(return_value=0),
                get_setting_as_float=Mock(return_value=0.0),
                get_setting_as_bool=Mock(return_value=True),
            ),
            "InputManager": Mock(
                sendKeys=Mock(),
                handle_device_grab=Mock(),
            ),
            "ScreenManager": Mock(update_screen_ignored=Mock()),
            "EventManager": Mock(stop_main_event_loop=Mock()),
            "MemoryManager": Mock(
                add_value_to_first_index=Mock(),
                get_index_list_element=Mock(return_value="test clipboard"),
                is_index_list_empty=Mock(return_value=False),
            ),
            "VmenuManager": Mock(
                set_curr_menu=Mock(),
            ),
            "CursorManager": Mock(
                set_window_for_application=Mock(),
                clear_window_for_application=Mock(),
            ),
        },
        "settings": Mock(),
        "general": {
            "curr_user": "testuser",
        },
    }
    return env


@pytest.fixture
def temp_config_file(tmp_path):
    """Create a temporary configuration file for testing.

    Returns path to a valid test configuration file.
    """
    config_path = tmp_path / "test_settings.conf"
    config_content = """[sound]
enabled=True
driver=gstreamerDriver
theme=default
volume=0.7

[speech]
enabled=True
driver=speechdDriver
rate=0.5
pitch=0.5
volume=1.0
autoReadIncoming=True

[screen]
driver=vcsaDriver
encoding=auto
screenUpdateDelay=0.05

[keyboard]
driver=evdevDriver
device=ALL
grabDevices=True
keyboardLayout=desktop

[general]
debugLevel=2
debugMode=File

[remote]
enable=True
driver=unixDriver
port=22447
enableSettingsRemote=True
enableCommandRemote=True
"""
    config_path.write_text(config_content)
    return config_path


@pytest.fixture
def temp_socket_path(tmp_path):
    """Create a temporary Unix socket path for testing.

    Returns path that can be used for Unix socket testing.
    """
    return tmp_path / "test_fenrir.sock"


@pytest.fixture
def temp_clipboard_file(tmp_path):
    """Create a temporary clipboard file for testing.

    Returns path to a temporary file for clipboard operations.
    """
    clipboard_path = tmp_path / "fenrirClipboard"
    clipboard_path.write_text("")
    return clipboard_path


@pytest.fixture
def sample_screen_data():
    """Return sample screen data for testing screen-related functionality.

    Returns dict with screen dimensions and content.
    """
    return {
        "columns": 80,
        "lines": 24,
        "delta": "Hello World",
        "cursor": {"x": 0, "y": 0},
        "content": "Sample screen content\nSecond line\nThird line",
    }


@pytest.fixture
def sample_remote_commands():
    """Return sample remote control commands for testing.

    Returns list of valid remote commands.
    """
    return [
        "command say Hello World",
        "command interrupt",
        "setting set speech#rate=0.8",
        "setting set speech#pitch=0.6",
        "setting set sound#volume=0.5",
        "setting reset",
    ]


@pytest.fixture
def invalid_remote_commands():
    """Return invalid remote control commands for testing validation.

    Returns list of commands that should be rejected.
    """
    return [
        "setting set speech#rate=999",  # Out of range
        "setting set speech#rate=-1",  # Negative value
        "setting set speech#pitch=10",  # Out of range
        "setting set speech#volume=-0.5",  # Negative volume
        "setting set invalid#setting=value",  # Invalid section
        "command unknown_command",  # Unknown command
    ]


# Pytest hooks for test session customization


def pytest_configure(config):
    """Configure pytest with custom settings."""
    # Add custom markers
    config.addinivalue_line(
        "markers", "unit: Unit tests (fast, no mocking required)"
    )
    config.addinivalue_line(
        "markers", "integration: Integration tests (require mocking)"
    )
    config.addinivalue_line(
        "markers", "driver: Driver tests (require root access)"
    )
    config.addinivalue_line(
        "markers", "slow: Tests that take more than 1 second"
    )


def pytest_collection_modifyitems(config, items):
    """Modify test collection to skip driver tests unless explicitly run.

    Driver tests require root access and hardware, so skip by default.
    Run with: pytest --run-driver-tests
    """
    skip_driver = pytest.mark.skip(
        reason="Driver tests require root access (use --run-driver-tests)"
    )
    run_driver_tests = config.getoption("--run-driver-tests", default=False)

    for item in items:
        if "driver" in item.keywords and not run_driver_tests:
            item.add_marker(skip_driver)


def pytest_addoption(parser):
    """Add custom command line options."""
    parser.addoption(
        "--run-driver-tests",
        action="store_true",
        default=False,
        help="Run driver tests that require root access",
    )



#!/bin/python
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import re
import string
import time

from fenrirscreenreader.core import debug
from fenrirscreenreader.core.i18n import _
from fenrirscreenreader.utils import line_utils


class OutputManager:
    def __init__(self):
        self.last_echo = ""

    def initialize(self, environment):
        self.env = environment
        self.env["runtime"]["SettingsManager"].load_driver(
            self.env["runtime"]["SettingsManager"].get_setting(
                "speech", "driver"
            ),
            "SpeechDriver",
        )
        self.env["runtime"]["SettingsManager"].load_driver(
            self.env["runtime"]["SettingsManager"].get_setting(
                "sound", "driver"
            ),
            "SoundDriver",
        )

    def shutdown(self):
        self.env["runtime"]["SettingsManager"].shutdown_driver("SoundDriver")
        self.env["runtime"]["SettingsManager"].shutdown_driver("SpeechDriver")

    def present_text(
        self,
        text,
        interrupt=True,
        sound_icon="",
        ignore_punctuation=False,
        announce_capital=False,
        flush=True,
    ):
        if text == "":
            return
        if (
            self.env["runtime"]["SettingsManager"].get_setting_as_bool(
                "speech", "readNumbersAsDigits"
            )
            and len(text.strip()) > 1
        ):
            text = re.sub(r"(\d)", r"\1 ", text).rstrip()
        self.env["runtime"]["DebugManager"].write_debug_out(
            "present_text:\nsoundIcon:'" + sound_icon + "'\nText:\n" + text,
            debug.DebugLevel.INFO,
        )
        if self.play_sound_icon(sound_icon, interrupt):
            self.env["runtime"]["DebugManager"].write_debug_out(
                "sound_icon found", debug.DebugLevel.INFO
            )
            return
        if (len(text) > 1) and (text.strip(string.whitespace) == ""):
            return
        to_announce_capital = announce_capital and text[0].isupper()
        if to_announce_capital:
            if self.play_sound_icon("capital", False):
                to_announce_capital = False
        self.last_echo = text
        self.speak_text(
            text, interrupt, ignore_punctuation, to_announce_capital, flush
        )

    def get_last_echo(self):
        return self.last_echo

    def process_mid_word_punctuation(self, text):
        """
        Process punctuation that appears mid-word to ensure proper pronunciation.
        Specifically handles dots between word characters (e.g., "settings.conf" -> "settings dot conf")
        and dots at word beginnings (e.g., ".local" -> "dot local")
        while preserving sentence-ending periods and other punctuation behavior.
        """
        if not text:
            return text
        
        # Handle dots at the beginning of words (like .local, .bashrc, .config)
        # Look for non-word character (or start of string), dot, then word characters
        text = re.sub(r'(?<!\w)\.(\w+)', r'dot \1', text)
        
        # Replace dots that appear between word characters with spoken form
        # Use a loop to handle multiple consecutive dots like www.example.com or a.b.c.d
        while re.search(r'\b\w+\.\w+\b', text):
            text = re.sub(r'\b(\w+)\.(\w+)\b', r'\1 dot \2', text)
        
        return text

    def speak_text(
        self,
        text,
        interrupt=True,
        ignore_punctuation=False,
        announce_capital=False,
        flush=True,
    ):
        if not self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "speech", "enabled"
        ):
            self.env["runtime"]["DebugManager"].write_debug_out(
                "Speech disabled in OutputManager.speak_text",
                debug.DebugLevel.INFO,
            )
            return
        if self.env["runtime"]["SpeechDriver"] is None:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "No SpeechDriver in OutputManager.speak_text",
                debug.DebugLevel.ERROR,
            )
            return
        if interrupt or flush:
            self.interrupt_output()
        try:
            self.env["runtime"]["SpeechDriver"].set_language(
                self.env["runtime"]["SettingsManager"].get_setting(
                    "speech", "language"
                )
            )
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "setting speech language in OutputManager.speak_text",
                debug.DebugLevel.ERROR,
            )
            self.env["runtime"]["DebugManager"].write_debug_out(
                str(e), debug.DebugLevel.ERROR
            )

        try:
            self.env["runtime"]["SpeechDriver"].set_voice(
                self.env["runtime"]["SettingsManager"].get_setting(
                    "speech", "voice"
                )
            )
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "Error while setting speech voice in OutputManager.speak_text",
                debug.DebugLevel.ERROR,
            )
            self.env["runtime"]["DebugManager"].write_debug_out(
                str(e), debug.DebugLevel.ERROR
            )

        try:
            if announce_capital:
                self.env["runtime"]["SpeechDriver"].set_pitch(
                    self.env["runtime"][
                        "SettingsManager"
                    ].get_setting_as_float("speech", "capitalPitch")
                )
            else:
                self.env["runtime"]["SpeechDriver"].set_pitch(
                    self.env["runtime"][
                        "SettingsManager"
                    ].get_setting_as_float("speech", "pitch")
                )
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "setting speech pitch in OutputManager.speak_text",
                debug.DebugLevel.ERROR,
            )
            self.env["runtime"]["DebugManager"].write_debug_out(
                str(e), debug.DebugLevel.ERROR
            )

        try:
            self.env["runtime"]["SpeechDriver"].set_rate(
                self.env["runtime"]["SettingsManager"].get_setting_as_float(
                    "speech", "rate"
                )
            )
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "setting speech rate in OutputManager.speak_text",
                debug.DebugLevel.ERROR,
            )
            self.env["runtime"]["DebugManager"].write_debug_out(
                str(e), debug.DebugLevel.ERROR
            )

        try:
            self.env["runtime"]["SpeechDriver"].set_module(
                self.env["runtime"]["SettingsManager"].get_setting(
                    "speech", "module"
                )
            )
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "setting speech module in OutputManager.speak_text",
                debug.DebugLevel.ERROR,
            )
            self.env["runtime"]["DebugManager"].write_debug_out(
                str(e), debug.DebugLevel.ERROR
            )

        try:
            self.env["runtime"]["SpeechDriver"].set_volume(
                self.env["runtime"]["SettingsManager"].get_setting_as_float(
                    "speech", "volume"
                )
            )
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "setting speech volume in OutputManager.speak_text ",
                debug.DebugLevel.ERROR,
            )
            self.env["runtime"]["DebugManager"].write_debug_out(
                str(e), debug.DebugLevel.ERROR
            )

        try:
            if self.env["runtime"]["SettingsManager"].get_setting_as_bool(
                "general", "newLinePause"
            ):
                clean_text = text.replace("\n", " , ")
            else:
                clean_text = text.replace("\n", " ")

            clean_text = self.env["runtime"]["TextManager"].replace_head_lines(
                clean_text
            )
            clean_text = self.process_mid_word_punctuation(clean_text)
            clean_text = self.env["runtime"][
                "PunctuationManager"
            ].proceed_punctuation(clean_text, ignore_punctuation)
            clean_text = re.sub(" +$", " ", clean_text)
            self.env["runtime"]["SpeechDriver"].speak(
                clean_text, True, ignore_punctuation
            )
            self.env["runtime"]["DebugManager"].write_debug_out(
                "Speak: " + clean_text, debug.DebugLevel.INFO
            )
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                '"speak" in OutputManager.speak_text ', debug.DebugLevel.ERROR
            )
            self.env["runtime"]["DebugManager"].write_debug_out(
                str(e), debug.DebugLevel.ERROR
            )

    def interrupt_output(self):
        try:
            self.env["runtime"]["SpeechDriver"].cancel()
            self.env["runtime"]["DebugManager"].write_debug_out(
                "Interrupt speech", debug.DebugLevel.INFO
            )
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "OutputManager interrupt_output: Error canceling speech: "
                + str(e),
                debug.DebugLevel.ERROR,
            )

    def play_sound_icon(self, sound_icon="", interrupt=True):
        if sound_icon == "":
            return False
        sound_icon = sound_icon.upper()
        if not self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "sound", "enabled"
        ):
            self.env["runtime"]["DebugManager"].write_debug_out(
                "Sound disabled in OutputManager.play_sound_icon",
                debug.DebugLevel.INFO,
            )
            return False

        try:
            e = self.env["soundIcons"][sound_icon]
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "OutputManager play_sound_icon: SoundIcon does not exist "
                + sound_icon
                + ": "
                + str(e),
                debug.DebugLevel.WARNING,
            )
            return False

        if self.env["runtime"]["SoundDriver"] is None:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "No SoundDriver in OutputManager.play_sound_icon:  SoundDriver not loaded",
                debug.DebugLevel.ERROR,
            )
            return False

        try:
            self.env["runtime"]["SoundDriver"].set_volume(
                self.env["runtime"]["SettingsManager"].get_setting_as_float(
                    "sound", "volume"
                )
            )
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "OutputManager.play_sound_icon::set_volume: " + str(e),
                debug.DebugLevel.ERROR,
            )

        try:
            self.env["runtime"]["SoundDriver"].play_sound_file(
                self.env["soundIcons"][sound_icon], interrupt
            )
            return True
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "OutputManager.play_sound_icon::play_sound_file: " + str(e),
                debug.DebugLevel.ERROR,
            )
            return False

        return False

    def play_frequence(self, frequence, duration, interrupt=True):
        if not self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "sound", "enabled"
        ):
            self.env["runtime"]["DebugManager"].write_debug_out(
                "Sound disabled in OutputManager.play_frequence",
                debug.DebugLevel.INFO,
            )
            return False

        if frequence < 1 or frequence > 20000:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "OutputManager.play_frequence::Filefrequence is out of range:"
                + str(frequence),
                debug.DebugLevel.INFO,
            )
            return False

        if self.env["runtime"]["SoundDriver"] is None:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "No SoundDriver in OutputManager.play_frequence: SoundDriver not loaded",
                debug.DebugLevel.ERROR,
            )
            return False

        try:
            self.env["runtime"]["SoundDriver"].set_volume(
                self.env["runtime"]["SettingsManager"].get_setting_as_float(
                    "sound", "volume"
                )
            )
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "OutputManager.play_sound_icon::set_volume: " + str(e),
                debug.DebugLevel.ERROR,
            )
        adjust_volume = 0.0
        try:
            adjust_volume = 1.0 - (frequence / 20000)
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "OutputManager play_frequence: Error calculating adjust volume: "
                + str(e),
                debug.DebugLevel.ERROR,
            )
        if adjust_volume > 9.0:
            adjust_volume = 9.0

        try:
            self.env["runtime"]["SoundDriver"].play_frequence(
                frequence, duration, adjust_volume, interrupt
            )
            return True
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "OutputManager.play_sound_icon::play_sound_file: " + str(e),
                debug.DebugLevel.ERROR,
            )
            return False

        return False

    def temp_disable_speech(self):
        if self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "speech", "enabled"
        ):
            self.present_text(
                _("speech temporary disabled"),
                sound_icon="SpeechOff",
                interrupt=True,
            )
            self.env["commandBuffer"]["enableSpeechOnKeypress"] = True
            # Also enable prompt watching for automatic speech restoration
            if "silenceUntilPrompt" not in self.env["commandBuffer"]:
                self.env["commandBuffer"]["silenceUntilPrompt"] = False
            self.env["commandBuffer"]["silenceUntilPrompt"] = True
            self.env["runtime"]["SettingsManager"].set_setting(
                "speech",
                "enabled",
                str(
                    not self.env["runtime"][
                        "SettingsManager"
                    ].get_setting_as_bool("speech", "enabled")
                ),
            )
            self.interrupt_output()

    def announce_active_cursor(self, interrupt_p=False):
        if self.env["runtime"]["CursorManager"].is_review_mode():
            self.present_text(" review cursor ", interrupt=interrupt_p)
        else:
            self.present_text(" text cursor ", interrupt=interrupt_p)

    def reset_SpeechDriver(self):
        """Reset speech driver to clean state - called by settings_manager"""
        if (
            "SpeechDriver" in self.env["runtime"]
            and self.env["runtime"]["SpeechDriver"]
        ):
            try:
                self.env["runtime"]["SpeechDriver"].reset()
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "Speech driver reset successfully", debug.DebugLevel.INFO
                )
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    f"reset_SpeechDriver error: {e}", debug.DebugLevel.ERROR
                )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import time
from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment
        # Initialize tab completion state tracking
        if "tabCompletion" not in self.env["commandBuffer"]:
            self.env["commandBuffer"]["tabCompletion"] = {
                "lastTabTime": 0,
                "pendingCompletion": None,
                "retryCount": 0
            }

    def shutdown(self):
        pass

    def get_description(self):
        return _("Announces tab completions when detected")

    def _is_recent_tab_input(self):
        """Check if TAB was pressed recently (within 200ms window)"""
        current_time = time.time()
        tab_detected = False
        
        # Check KEY mode
        if self.env["runtime"]["InputManager"].get_shortcut_type() in ["KEY"]:
            if (self.env["runtime"]["InputManager"].get_last_deepest_input()
                in [["KEY_TAB"]]):
                tab_detected = True
                self.env["commandBuffer"]["tabCompletion"]["lastTabTime"] = current_time
        
        # Check BYTE mode
        elif self.env["runtime"]["InputManager"].get_shortcut_type() in ["BYTE"]:
            for currByte in self.env["runtime"]["ByteManager"].get_last_byte_key():
                if currByte == 9:  # Tab character
                    tab_detected = True
                    self.env["commandBuffer"]["tabCompletion"]["lastTabTime"] = current_time
        
        # Check if tab was pressed recently (200ms window)
        if not tab_detected:
            time_since_tab = current_time - self.env["commandBuffer"]["tabCompletion"]["lastTabTime"]
            if time_since_tab <= 0.2:  # 200ms window
                tab_detected = True
        
        return tab_detected

    def _is_flexible_completion_match(self, x_move, delta_text):
        """Use flexible matching instead of strict equality"""
        if not delta_text:
            return False
        
        delta_len = len(delta_text)
        
        # Exact match (preserve original behavior)
        if x_move == delta_len:
            return True
        
        # Flexible range: allow ±2 characters difference
        # Handles spacing adjustments and unicode width variations
        if abs(x_move - delta_len) <= 2 and delta_len > 0:
            return True
        
        # For longer completions, allow proportional variance
        if delta_len > 10 and abs(x_move - delta_len) <= (delta_len * 0.2):
            return True
        
        return False

    def _detect_completion_patterns(self, delta_text):
        """Detect common tab completion patterns for improved accuracy"""
        if not delta_text:
            return False
        
        delta_stripped = delta_text.strip()
        
        # File extension completion
        if '.' in delta_stripped and delta_stripped.count('.') <= 2:
            return True
        
        # Path completion (contains / or \)
        if '/' in delta_stripped or '\\' in delta_stripped:
            return True
        
        # Command parameter completion (starts with -)
        if delta_stripped.startswith('-') and len(delta_stripped) > 1:
            return True
        
        # Word boundary completion (alphanumeric content)
        if delta_stripped.isalnum() and len(delta_stripped) >= 2:
            return True
        
        return False

    def run(self):
        """Enhanced tab completion detection with improved reliability"""
        # Basic cursor movement check (preserve original logic)
        x_move = (
            self.env["screen"]["new_cursor"]["x"]
            - self.env["screen"]["old_cursor"]["x"]
        )
        if x_move <= 0:
            return

        # Enhanced tab input detection with persistence
        tab_detected = self._is_recent_tab_input()
        
        # Fallback for non-tab movements (preserve original thresholds)
        if not tab_detected:
            if x_move < 5:
                return

        # Screen delta availability check
        if not self.env["runtime"]["ScreenManager"].is_delta():
            # If tab was detected but no delta yet, store for potential retry
            if tab_detected and self.env["commandBuffer"]["tabCompletion"]["retryCount"] < 2:
                self.env["commandBuffer"]["tabCompletion"]["pendingCompletion"] = {
                    "x_move": x_move,
                    "timestamp": time.time()
                }
                self.env["commandBuffer"]["tabCompletion"]["retryCount"] += 1
            return

        delta_text = self.env["screen"]["new_delta"]
        
        # Enhanced correlation checking with flexible matching
        if not self._is_flexible_completion_match(x_move, delta_text):
            # Additional pattern-based validation for edge cases
            if not (tab_detected and self._detect_completion_patterns(delta_text)):
                return

        # Reset retry counter on successful detection
        self.env["commandBuffer"]["tabCompletion"]["retryCount"] = 0
        self.env["commandBuffer"]["tabCompletion"]["pendingCompletion"] = None

        # Mark that we've handled this delta to prevent duplicate announcements
        # This prevents the incoming text handler from also announcing the same content
        self.env["commandBuffer"]["tabCompletion"]["lastProcessedDelta"] = delta_text
        self.env["commandBuffer"]["tabCompletion"]["lastProcessedTime"] = time.time()

        # Text filtering and announcement (preserve original behavior)
        curr_delta = delta_text
        if (len(curr_delta.strip()) != len(curr_delta) and curr_delta.strip() != ""):
            curr_delta = curr_delta.strip()
        
        # Enhanced announcement with better handling of empty completions
        if curr_delta:
            self.env["runtime"]["OutputManager"].present_text(
                curr_delta, interrupt=True, announce_capital=True, flush=False
            )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3

import importlib.util
import os

from fenrirscreenreader.core.i18n import _

# Load base configuration class
_base_path = os.path.join(os.path.dirname(__file__), "..", "config_base.py")
_spec = importlib.util.spec_from_file_location("config_base", _base_path)
_module = importlib.util.module_from_spec(_spec)
_spec.loader.exec_module(_module)
config_command = _module.config_command


class command(config_command):
    def __init__(self):
        super().__init__()

    def get_description(self):
        return "Select sound theme"

    def run(self):
        current_theme = self.get_setting("sound", "theme", "default")

        # Present current theme
        self.present_text(f"Current sound theme: {current_theme}")

        # Look for available sound themes
        sound_paths = [
            "/usr/share/sounds",
            "/usr/share/fenrirscreenreader/sounds",
            os.path.expanduser("~/.local/share/fenrirscreenreader/sounds"),
        ]

        available_themes = self.get_available_themes(sound_paths)

        if len(available_themes) > 1:
            # For this implementation, cycle through available themes
            try:
                current_index = available_themes.index(current_theme)
                next_index = (current_index + 1) % len(available_themes)
                new_theme = available_themes[next_index]
            except ValueError:
                # Current theme not found in available themes, use first one
                new_theme = available_themes[0]

            success = self.set_setting("sound", "theme", new_theme)

            if success:
                self.present_text(f"Sound theme changed to: {new_theme}")
                self.play_sound("Accept")
            else:
                self.present_text("Failed to change sound theme")
                self.play_sound("Error")
        else:
            self.present_text("Only default sound theme is available")
            self.play_sound("Cancel")

    def get_available_themes(self, searchPaths):
        """Find available sound themes in the search paths"""
        themes = ["default"]  # Always include default

        for path in searchPaths:
            if os.path.isdir(path):
                try:
                    for item in os.listdir(path):
                        full_path = os.path.join(path, item)
                        if (
                            os.path.isdir(full_path)
                            and item != "default"
                            and item not in themes
                        ):
                            # Check if it looks like a sound theme (contains
                            # sound files)
                            if self.is_valid_sound_theme(full_path):
                                themes.append(item)
                except Exception:
                    continue

        return themes

    def is_valid_sound_theme(self, themePath):
        """Check if a directory contains sound files"""
        sound_extensions = [".wav", ".ogg", ".mp3", ".flac"]
        try:
            for file in os.listdir(themePath):
                if any(file.lower().endswith(ext) for ext in sound_extensions):
                    return True
        except Exception:
            pass
        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import glob
import inspect
import os
import time

from fenrirscreenreader.core import debug
from fenrirscreenreader.utils import module_utils

currentdir = os.path.dirname(
    os.path.realpath(os.path.abspath(inspect.getfile(inspect.currentframe())))
)
fenrir_path = os.path.dirname(currentdir)


class CommandManager:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment
        # commands
        self.env["commands"] = {}
        self.env["commandsIgnore"] = {}
        for command_folder in self.env["general"]["commandFolderList"]:
            self.env["runtime"]["CommandManager"].load_commands(command_folder)
            if (
                self.env["runtime"]["SettingsManager"].get_setting(
                    "general", "command_path"
                )
                != ""
            ):
                self.env["runtime"]["CommandManager"].load_commands(
                    command_folder,
                    self.env["runtime"]["SettingsManager"].get_setting(
                        "general", "command_path"
                    ),
                )

        # scripts for scriptKey
        self.env["runtime"]["CommandManager"].load_script_commands()

    def shutdown(self):
        for command_folder in self.env["general"]["commandFolderList"]:
            self.env["runtime"]["CommandManager"].shutdown_commands(
                command_folder
            )

    def load_file(self, filepath=""):
        if filepath == "":
            return None
        if not os.path.exists(filepath):
            self.env["runtime"]["DebugManager"].write_debug_out(
                "load_file: filepath not exists:" + filepath,
                debug.DebugLevel.WARNING,
            )
            return None
        if os.path.isdir(filepath):
            self.env["runtime"]["DebugManager"].write_debug_out(
                "load_file: filepath is a directory:" + filepath,
                debug.DebugLevel.ERROR,
            )
            return None
        if not os.access(filepath, os.R_OK):
            self.env["runtime"]["DebugManager"].write_debug_out(
                "load_file: filepath not readable:" + filepath,
                debug.DebugLevel.ERROR,
            )
            return None

        try:
            file_name, file_extension = os.path.splitext(filepath)
            file_name = file_name.split("/")[-1]
            if file_name.startswith("__"):
                return None
            if file_extension.lower() == ".py":
                command_mod = module_utils.import_module(file_name, filepath)
                command = command_mod.command()
                command.initialize(self.env)
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "load_file: Load command:" + filepath,
                    debug.DebugLevel.INFO,
                    on_any_level=True,
                )
                return command
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "load_file: Loading command:" + filepath,
                debug.DebugLevel.ERROR,
            )
            self.env["runtime"]["DebugManager"].write_debug_out(
                str(e), debug.DebugLevel.ERROR
            )
        return None

    def load_commands(self, section="commands", command_path=""):
        if command_path == "":
            command_path = fenrir_path + "/commands/"
        if not command_path.endswith("/"):
            command_path += "/"
        command_folder = command_path + section + "/"
        if not os.path.exists(command_folder):
            self.env["runtime"]["DebugManager"].write_debug_out(
                "load_commands: command_folder not exists:" + command_folder,
                debug.DebugLevel.WARNING,
            )
            return
        if not os.path.isdir(command_folder):
            self.env["runtime"]["DebugManager"].write_debug_out(
                "load_commands: command_folder not a directory:"
                + command_folder,
                debug.DebugLevel.ERROR,
            )
            return
        if not os.access(command_folder, os.R_OK):
            self.env["runtime"]["DebugManager"].write_debug_out(
                "load_commands: command_folder not readable:" + command_folder,
                debug.DebugLevel.ERROR,
            )
            return
        self.env["commands"][section] = {}
        self.env["commandsIgnore"][section] = {}
        command_list = glob.glob(command_folder + "*")
        for command in command_list:
            try:
                file_name, file_extension = os.path.splitext(command)
                file_name = file_name.split("/")[-1]
                if file_name.startswith("__"):
                    continue
                # Skip base classes that shouldn't be loaded as commands
                if file_name.endswith("_base"):
                    continue
                # Check if command already exists to prevent duplicate loading
                if (
                    file_name.upper() in self.env["commands"][section]
                    and self.env["commands"][section][file_name.upper()]
                    is not None
                ):
                    continue
                if file_extension.lower() == ".py":
                    command_mod = module_utils.import_module(
                        file_name, command
                    )
                    self.env["commands"][section][
                        file_name.upper()
                    ] = command_mod.command()
                    self.env["commandsIgnore"][section][
                        file_name.upper()[file_name.upper().find("-") + 1 :]
                        + "_IGNORE"
                    ] = False
                    self.env["commands"][section][
                        file_name.upper()
                    ].initialize(self.env)
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        "load_commands: Load command:"
                        + section
                        + "."
                        + file_name.upper(),
                        debug.DebugLevel.INFO,
                        on_any_level=True,
                    )
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "load_commands: Loading command:" + command,
                    debug.DebugLevel.ERROR,
                )
                self.env["runtime"]["DebugManager"].write_debug_out(
                    str(e), debug.DebugLevel.ERROR
                )
                continue

    def load_script_commands(self, section="commands", script_path=""):
        if script_path == "":
            script_path = self.env["runtime"]["SettingsManager"].get_setting(
                "general", "script_path"
            )
        if not script_path.endswith("/"):
            script_path += "/"
        if not os.path.exists(script_path):
            if os.path.exists(fenrir_path + "/../../config/scripts/"):
                script_path = fenrir_path + "/../../config/scripts/"
            else:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "scriptpath not exists:" + script_path,
                    debug.DebugLevel.WARNING,
                )
                return
        if not os.path.isdir(script_path):
            self.env["runtime"]["DebugManager"].write_debug_out(
                "scriptpath not a directory:" + script_path,
                debug.DebugLevel.ERROR,
            )
            return
        if not os.access(script_path, os.R_OK):
            self.env["runtime"]["DebugManager"].write_debug_out(
                "scriptpath not readable:" + script_path,
                debug.DebugLevel.ERROR,
            )
            return
        command_list = glob.glob(script_path + "*")
        sub_command = fenrir_path + "/commands/commands/subprocess.py"
        for command in command_list:
            invalid = False
            try:
                file_name, file_extension = os.path.splitext(command)
                file_name = file_name.split("/")[-1]
                if file_name.startswith("__"):
                    continue
                if file_name.upper() in self.env["commands"][section]:
                    continue
                command_mod = module_utils.import_module(
                    file_name, sub_command
                )
                self.env["commands"][section][
                    file_name.upper()
                ] = command_mod.command()
                self.env["commands"][section][file_name.upper()].initialize(
                    self.env, command
                )
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "Load script:" + section + "." + file_name.upper(),
                    debug.DebugLevel.INFO,
                    on_any_level=True,
                )
                comm_settings = file_name.upper().split("__-__")
                if len(comm_settings) == 1:
                    keys = comm_settings[0]
                elif len(comm_settings) == 2:
                    keys = comm_settings[1]
                elif len(comm_settings) > 2:
                    continue
                keys = keys.split("__+__")
                shortcut_keys = []
                shortcut = []
                for key in keys:
                    if not self.env["runtime"]["InputManager"].is_valid_key(
                        key.upper()
                    ):
                        self.env["runtime"]["DebugManager"].write_debug_out(
                            "invalid key : "
                            + key.upper()
                            + " script:"
                            + file_name,
                            debug.DebugLevel.WARNING,
                        )
                        invalid = True
                        break
                    shortcut_keys.append(key.upper())
                if invalid:
                    continue
                if "KEY_SCRIPT" not in shortcut_keys:
                    shortcut_keys.append("KEY_SCRIPT")
                shortcut.append(1)
                shortcut.append(sorted(shortcut_keys))
                self.env["bindings"][str(shortcut)] = file_name.upper()
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "Loading script:" + file_name, debug.DebugLevel.ERROR
                )
                self.env["runtime"]["DebugManager"].write_debug_out(
                    str(e), debug.DebugLevel.ERROR
                )
                continue

    def shutdown_commands(self, section):
        # Check if the section exists in the commands dictionary
        if section not in self.env["commands"]:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "shutdown_commands: section not found:" + section,
                debug.DebugLevel.WARNING,
            )
            return

        for command in sorted(self.env["commands"][section]):
            try:
                self.env["commands"][section][command].shutdown()
                del self.env["commands"][section][command]
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "Shutdown command:" + section + "." + command,
                    debug.DebugLevel.ERROR,
                )
                self.env["runtime"]["DebugManager"].write_debug_out(
                    str(e), debug.DebugLevel.ERROR
                )
                continue

    def execute_switch_trigger(self, trigger, unLoadScript, loadScript):
        if self.env["runtime"]["ScreenManager"].is_ignored_screen():
            return
        # unload
        old_script = unLoadScript
        if self.command_exists(old_script, trigger):
            try:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "Executing switchtrigger.unload:"
                    + trigger
                    + "."
                    + old_script,
                    debug.DebugLevel.INFO,
                )
                self.env["commands"][trigger][old_script].unload()
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "Executing trigger:" + trigger + "." + old_script,
                    debug.DebugLevel.ERROR,
                )
                self.env["runtime"]["DebugManager"].write_debug_out(
                    str(e), debug.DebugLevel.ERROR
                )
        # load
        new_script = loadScript
        if self.command_exists(new_script, trigger):
            try:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "Executing switchtrigger.load:"
                    + trigger
                    + "."
                    + new_script,
                    debug.DebugLevel.INFO,
                )
                self.env["commands"][trigger][new_script].load()
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "Executing trigger:" + trigger + "." + new_script,
                    debug.DebugLevel.ERROR,
                )
                self.env["runtime"]["DebugManager"].write_debug_out(
                    str(e), debug.DebugLevel.ERROR
                )

    def execute_default_trigger(self, trigger, force=False):
        if not force:
            if self.env["runtime"]["ScreenManager"].is_ignored_screen():
                return
        for command in sorted(self.env["commands"][trigger]):
            if self.command_exists(command, trigger):
                try:
                    if self.env["commandsIgnore"][trigger][
                        command[command.find("-") + 1 :] + "_IGNORE"
                    ]:
                        self.env["commandsIgnore"][trigger][
                            command[command.find("-") + 1 :] + "_IGNORE"
                        ] = False
                        self.env["runtime"]["DebugManager"].write_debug_out(
                            "ignore trigger.command:"
                            + trigger
                            + "."
                            + command,
                            debug.DebugLevel.INFO,
                        )
                    else:
                        self.env["runtime"]["DebugManager"].write_debug_out(
                            "Executing trigger.command:"
                            + trigger
                            + "."
                            + command,
                            debug.DebugLevel.INFO,
                        )
                        self.env["commands"][trigger][command].run()
                except Exception as e:
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        "Executing trigger:"
                        + trigger
                        + "."
                        + command
                        + str(e),
                        debug.DebugLevel.ERROR,
                    )

    def execute_command(self, command, section="commands"):
        if self.env["runtime"]["ScreenManager"].is_ignored_screen():
            return
        if self.command_exists(command, section):
            try:
                if (
                    self.env["runtime"]["HelpManager"].is_tutorial_mode()
                    and section != "help"
                ):
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        "Tutorial for command:" + section + "." + command,
                        debug.DebugLevel.INFO,
                    )
                    description = self.get_command_description(
                        command, section
                    )
                    self.env["runtime"]["OutputManager"].present_text(
                        description, interrupt=False
                    )
                else:
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        "Executing command:" + section + "." + command,
                        debug.DebugLevel.INFO,
                    )
                    self.run_command(command, section)
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "Executing command:"
                    + section
                    + "."
                    + command
                    + " "
                    + str(e),
                    debug.DebugLevel.ERROR,
                )

    def run_command(self, command, section="commands"):
        if self.command_exists(command, section):
            try:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "run_command command:" + section + "." + command,
                    debug.DebugLevel.INFO,
                )
                self.env["commands"][section][command].run()
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "run_command command:"
                    + section
                    + "."
                    + command
                    + " "
                    + str(e),
                    debug.DebugLevel.ERROR,
                )
        self.env["commandInfo"]["lastCommandExecutionTime"] = time.time()

    def get_command_description(self, command, section="commands"):
        if self.command_exists(command, section):
            try:
                return self.env["commands"][section][command].get_description()
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "command_manager.get_command_description:" + str(e),
                    debug.DebugLevel.ERROR,
                )
        self.env["commandInfo"]["lastCommandExecutionTime"] = time.time()

    def command_exists(self, command, section="commands"):
        try:
            return command in self.env["commands"][section]
        except Exception as e:
            return False

    def get_shortcut_for_command(self, command, formatKeys=False):
        shortcut = []
        try:
            raw_shortcut = list(self.env["bindings"].keys())[
                list(self.env["bindings"].values()).index(command)
            ]
            # prefer numbers for multitap
            for k in ["2", "3", "4", "5", "6", "7", "8"]:
                if k in raw_shortcut:
                    formatted_key = k
                    if formatKeys:
                        formatted_key = formatted_key.lower()
                        formatted_key += " times "
                    shortcut.append(formatted_key)
                    raw_shortcut.remove(k)
            # prefer metha keys
            for k in [
                "KEY_FENRIR",
                "KEY_SCRIPT",
                "KEY_CTRL",
                "KEY_SHIFT",
                "KEY_ALT",
                "KEY_META",
            ]:
                if k in raw_shortcut:
                    formatted_key = k
                    if formatKeys:
                        formatted_key = formatted_key.lower()
                        formatted_key = formatted_key.replace(
                            "key_kp", " keypad "
                        )
                        formatted_key = formatted_key.replace("key_", " ")
                    shortcut.append(formatted_key)
                    raw_shortcut.remove(k)
            # handle other keys
            for k in raw_shortcut:
                formatted_key = k
                if formatKeys:
                    formatted_key = formatted_key.lower()
                    formatted_key = formatted_key.replace("key_kp", " keypad ")
                    formatted_key = formatted_key.replace("key_", " ")
                shortcut.append(formatted_key)
        except Exception as e:
            pass
        return shortcut



#!/usr/bin/env python3

import time

from fenrirscreenreader.core.i18n import _

# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return "No Description found"

    def run(self):
        if not self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "promote", "enabled"
        ):
            return
        if (
            self.env["runtime"]["SettingsManager"]
            .get_setting("promote", "list")
            .strip(" \t\n")
            == ""
        ):
            return
        if int(time.time() - self.env["input"]["lastInputTime"]) < self.env[
            "runtime"
        ]["SettingsManager"].get_setting_as_int(
            "promote", "inactiveTimeoutSec"
        ):
            return
        if (
            len(
                self.env["runtime"]["SettingsManager"].get_setting(
                    "promote", "list"
                )
            )
            == 0
        ):
            return
        for promote in (
            self.env["runtime"]["SettingsManager"]
            .get_setting("promote", "list")
            .split(",")
        ):
            if promote in self.env["screen"]["new_delta"]:
                self.env["runtime"]["OutputManager"].play_sound_icon(
                    "PromotedText"
                )
                self.env["input"]["lastInputTime"] = time.time()
                return

    def set_callback(self, callback):
        pass
